"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_extra_1 = __importDefault(require("fs-extra"));
const os_1 = __importDefault(require("os"));
const path_1 = __importDefault(require("path"));
const file_box_1 = require("file-box");
const flash_store_1 = require("flash-store");
const rx_queue_1 = require("rx-queue");
const state_switch_1 = require("state-switch");
const wechaty_puppet_1 = require("wechaty-puppet");
const padchat_schemas_1 = require("./padchat-schemas");
const padchat_rpc_type_1 = require("./padchat-rpc.type");
const padchat_rpc_1 = require("./padchat-rpc");
const pure_function_helpers_1 = require("./pure-function-helpers/");
const config_1 = require("./config");
const MEMORY_SLOT_NAME = 'WECHATY_PUPPET_PADCHAT';
class PadchatManager extends padchat_rpc_1.PadchatRpc {
    constructor(options) {
        super(options.endpoint, options.token);
        this.options = options;
        config_1.log.verbose('PuppetPadchatManager', 'constructor()');
        this.memorySlot = {
            device: {},
        };
        this.state = new state_switch_1.StateSwitch('PuppetPadchatManager');
        /**
         * Executer Queue: execute one task at a time,
         *  delay between them for 15 second
         */
        this.delayQueueExecutor = new rx_queue_1.DelayQueueExector(1000 * 15);
        this.contactListSynced = false;
        this.roomNeedsToBeSync = 0;
    }
    initCache(token, userId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', 'initCache(%s, %s)', token, userId);
            if (this.cacheContactRawPayload
                || this.cacheRoomMemberRawPayload
                || this.cacheRoomRawPayload
                || this.cacheRoomInvitationRawPayload) {
                throw new Error('cache exists');
            }
            const baseDir = path_1.default.join(os_1.default.homedir(), path_1.default.sep, '.wechaty', 'puppet-padchat-cache', path_1.default.sep, token, path_1.default.sep, userId);
            const baseDirExist = yield fs_extra_1.default.pathExists(baseDir);
            if (!baseDirExist) {
                yield fs_extra_1.default.mkdirp(baseDir);
            }
            this.cacheContactRawPayload = new flash_store_1.FlashStoreSync(path_1.default.join(baseDir, 'contact-raw-payload'));
            this.cacheRoomMemberRawPayload = new flash_store_1.FlashStoreSync(path_1.default.join(baseDir, 'room-member-raw-payload'));
            this.cacheRoomRawPayload = new flash_store_1.FlashStoreSync(path_1.default.join(baseDir, 'room-raw-payload'));
            this.cacheRoomInvitationRawPayload = new flash_store_1.FlashStoreSync(path_1.default.join(baseDir, 'room-invitation-raw-payload'));
            yield Promise.all([
                this.cacheContactRawPayload.ready(),
                this.cacheRoomMemberRawPayload.ready(),
                this.cacheRoomRawPayload.ready(),
                this.cacheRoomInvitationRawPayload.ready(),
            ]);
            const roomMemberTotalNum = [...this.cacheRoomMemberRawPayload.values()].reduce((accuVal, currVal) => {
                return accuVal + Object.keys(currVal).length;
            }, 0);
            config_1.log.verbose('PuppetPadchatManager', 'initCache() inited %d Contacts, %d RoomMembers, %d Rooms, cachedir="%s"', this.cacheContactRawPayload.size, roomMemberTotalNum, this.cacheRoomRawPayload.size, baseDir);
        });
    }
    releaseCache() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', 'releaseCache()');
            if (this.cacheContactRawPayload
                && this.cacheRoomMemberRawPayload
                && this.cacheRoomRawPayload
                && this.cacheRoomInvitationRawPayload) {
                config_1.log.silly('PuppetPadchatManager', 'releaseCache() closing caches ...');
                yield Promise.all([
                    this.cacheContactRawPayload.close(),
                    this.cacheRoomMemberRawPayload.close(),
                    this.cacheRoomRawPayload.close(),
                    this.cacheRoomInvitationRawPayload.close(),
                ]);
                this.cacheContactRawPayload = undefined;
                this.cacheRoomMemberRawPayload = undefined;
                this.cacheRoomRawPayload = undefined;
                this.cacheRoomInvitationRawPayload = undefined;
                config_1.log.silly('PuppetPadchatManager', 'releaseCache() cache closed.');
            }
            else {
                config_1.log.verbose('PuppetPadchatManager', 'releaseCache() cache not exist.');
            }
        });
    }
    start() {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `start()`);
            if (this.userId) {
                throw new Error('userId exist');
            }
            this.state.on('pending');
            /**
             * Sometimes the RPC WebSocket will failure on connect in super.start(),
             *  if that's true then a Error will be throw out.
             *  Try again in the following while loop until the state is not on('pending')
             */
            while (this.state.on() === 'pending') {
                try {
                    yield _super("start").call(this);
                    break;
                }
                catch (e) {
                    config_1.log.warn('PuppetPadchatManager', 'start() super.start() exception: %s', e);
                    yield _super("stop").call(this);
                    yield new Promise(r => setTimeout(r, 1000));
                    config_1.log.warn('PuppetPadchatManager', 'start() super.start() retry now ...');
                }
            }
            if (this.delayQueueExecutorSubscription) {
                throw new Error('this.delayExecutorSubscription exist');
            }
            else {
                this.delayQueueExecutorSubscription = this.delayQueueExecutor.subscribe(unit => {
                    config_1.log.verbose('PuppetPadchatManager', 'startQueues() delayQueueExecutor.subscribe(%s) executed', unit.name);
                    if (this.roomNeedsToBeSync === 0 && this.contactListSynced) {
                        this.emit('ready');
                    }
                    else {
                        config_1.log.verbose('PuppetPadchatManager', 'startQueues() delayQueueExecutor.subscribe(%s) %d rooms need to be synced', unit.name, this.roomNeedsToBeSync);
                    }
                });
            }
            this.memorySlot = Object.assign({}, this.memorySlot, yield this.options.memory.get(MEMORY_SLOT_NAME));
            yield this.tryLoad62Data();
            const succeed = yield this.tryAutoLogin(this.memorySlot);
            if (!succeed) {
                yield this.startCheckScan();
            }
            this.state.on(true);
        });
    }
    reconnect() {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'reconnect(), retry attempt left: %d', this.connectionStatus.reconnectLeft);
            this.updateConnectionStatus(padchat_rpc_1.CONNECTING);
            try {
                yield _super("reconnect").call(this);
                yield this.tryLoad62Data();
                yield this.tryAutoLogin(this.memorySlot);
                this.memorySlot = yield this.refreshMemorySlotData(this.memorySlot, this.userId);
                yield this.options.memory.set(MEMORY_SLOT_NAME, this.memorySlot);
                yield this.options.memory.save();
                this.updateConnectionStatus(padchat_rpc_1.CONNECTED);
            }
            catch (e) {
                if (this.connectionStatus.reconnectLeft === 0) {
                    this.updateConnectionStatus(padchat_rpc_1.DISCONNECTED);
                    this.emit('error', new Error('Can not connect to wechaty-puppet-padchat server'));
                }
                else {
                    config_1.log.verbose('PuppetPadchat', 'reconnect(), failed to reconnect, retrying in %d seconds', this.connectionStatus.interval / 1000);
                    this.cleanConnection();
                    yield new Promise(r => setTimeout(r, this.connectionStatus.interval));
                    return this.reconnect();
                }
            }
        });
    }
    stop() {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `stop()`);
            this.state.off('pending');
            if (this.delayQueueExecutorSubscription) {
                this.delayQueueExecutorSubscription.unsubscribe();
                this.delayQueueExecutorSubscription = undefined;
            }
            else {
                config_1.log.warn('PuppetPadchatManager', 'stop() subscript not exist');
            }
            yield this.stopCheckScan();
            yield _super("stop").call(this);
            yield this.releaseCache();
            this.userId = undefined;
            this.loginScanQrcode = undefined;
            this.loginScanStatus = undefined;
            this.contactListSynced = false;
            this.roomNeedsToBeSync = 0;
            this.state.off(true);
        });
    }
    onLogin(userId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `login(%s)`, userId);
            if (this.userId) {
                config_1.log.verbose('PuppetPadchatManager', 'reconnected(%s)', userId);
                return;
            }
            this.userId = userId;
            yield this.stopCheckScan();
            /**
             * Update Memory Slot
             */
            this.memorySlot = yield this.refreshMemorySlotData(this.memorySlot, userId);
            yield this.options.memory.set(MEMORY_SLOT_NAME, this.memorySlot);
            yield this.options.memory.save();
            /**
             * Init persistence cache
             */
            yield this.initCache(this.options.token, this.userId);
            /**
             * Refresh the login-ed user payload
             */
            if (this.cacheContactRawPayload) {
                this.cacheContactRawPayload.delete(this.userId);
                yield this.contactRawPayload(this.userId);
            }
            this.updateConnectionStatus(padchat_rpc_1.CONNECTED);
            this.emit('login', this.userId);
        });
    }
    logout() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `logout()`);
            if (!this.userId) {
                config_1.log.warn('PuppetPadchatManager', 'logout() userId not exist, already logout-ed');
                return;
            }
            if (this.delayQueueExecutorSubscription) {
                this.delayQueueExecutorSubscription.unsubscribe();
                this.delayQueueExecutorSubscription = undefined;
            }
            else {
                config_1.log.warn('PuppetPadchatManager', 'logout() subscript not exist');
            }
            this.userId = undefined;
            yield this.releaseCache();
            yield this.startCheckScan();
        });
    }
    stopCheckScan() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `stopCheckScan()`);
            if (this.loginScanTimer) {
                clearTimeout(this.loginScanTimer);
                this.loginScanTimer = undefined;
            }
            this.loginScanQrcode = undefined;
            this.loginScanStatus = undefined;
        });
    }
    startCheckScan() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `startCheckScan()`);
            if (this.userId) {
                config_1.log.warn('PuppetPadchatManager', 'startCheckScan() this.userId exist.');
                yield this.onLogin(this.userId);
                return;
            }
            if (this.loginScanTimer) {
                config_1.log.warn('PuppetPadchatManager', 'startCheckScan() this.loginScanTimer exist.');
                return;
            }
            const checkScanInternalLoop = () => __awaiter(this, void 0, void 0, function* () {
                config_1.log.silly('PuppetPadchatManager', `startCheckScan() checkScanInternalLoop()`);
                /**
                 * While we want to Wait user response
                 */
                let waitUserResponse = true;
                while (waitUserResponse) {
                    // const result = await this.padchatRpc.WXCheckQRCode()
                    const result = yield this.WXCheckQRCode();
                    if (this.loginScanStatus !== result.status && this.loginScanQrcode) {
                        this.loginScanStatus = result.status;
                        this.emit('scan', this.loginScanQrcode, this.loginScanStatus);
                    }
                    if (result.expired_time && result.expired_time < 10) {
                        /**
                         * result.expire_time is second
                         * emit new qrcode 10 seconds before the old one expired
                         */
                        this.loginScanQrcode = undefined;
                        this.loginScanStatus = undefined;
                        waitUserResponse = false;
                        continue;
                    }
                    switch (result.status) {
                        case padchat_rpc_type_1.WXCheckQRCodeStatus.WaitScan:
                            config_1.log.silly('PuppetPadchatManager', 'checkQrcode: Please scan the Qrcode!');
                            break;
                        case padchat_rpc_type_1.WXCheckQRCodeStatus.WaitConfirm:
                            /**
                             * WXCheckQRCode result:
                             * {
                             *  "expired_time": 236,
                             *  "head_url": "http://wx.qlogo.cn/mmhead/ver_1/5VaXXlAx53wb3M46gQpVtLiaVMd4ezhxOibJiaZXLf2ajTNPZloJI7QEpVxd4ibgpEnLF8gHVuLricaJesjJpsFiciaOw/0",
                             *  "nick_name": "李卓桓",
                             *  "status": 1
                             * }
                             */
                            config_1.log.silly('PuppetPadchatManager', 'checkQrcode: Had scan the Qrcode, but not Login!');
                            break;
                        case padchat_rpc_type_1.WXCheckQRCodeStatus.Confirmed:
                            config_1.log.silly('PuppetPadchatManager', 'checkQrcode: Trying to login... please wait');
                            if (!result.user_name || !result.password) {
                                throw Error('PuppetPadchatManager, checkQrcode, cannot get username or password here, return!');
                            }
                            const loginResult = yield this.WXQRCodeLogin(result.user_name, result.password);
                            yield this.onLogin(loginResult.user_name);
                            return;
                        case padchat_rpc_type_1.WXCheckQRCodeStatus.Timeout:
                            config_1.log.silly('PuppetPadchatManager', 'checkQrcode: Timeout');
                            this.loginScanQrcode = undefined;
                            this.loginScanStatus = undefined;
                            waitUserResponse = false;
                            break;
                        case padchat_rpc_type_1.WXCheckQRCodeStatus.Cancel:
                            config_1.log.silly('PuppetPadchatManager', 'user cancel');
                            this.loginScanQrcode = undefined;
                            this.loginScanStatus = undefined;
                            waitUserResponse = false;
                            break;
                        case padchat_rpc_type_1.WXCheckQRCodeStatus.Ignore:
                            config_1.log.silly('PuppetPadchatManager', 'ignore status -2');
                            this.loginScanQrcode = undefined;
                            this.loginScanStatus = undefined;
                            waitUserResponse = false;
                            break;
                        default:
                            config_1.log.warn('PuppetPadchatManager', 'startCheckScan() unknown WXCheckQRCodeStatus: ' + result.status);
                            this.loginScanQrcode = undefined;
                            this.loginScanStatus = undefined;
                            waitUserResponse = false;
                            break;
                    }
                    yield new Promise(r => setTimeout(r, 1000));
                }
                yield this.emitLoginQrcode();
                this.loginScanTimer = setTimeout(() => __awaiter(this, void 0, void 0, function* () {
                    this.loginScanTimer = undefined;
                    yield checkScanInternalLoop();
                }), 1000);
                return;
            });
            checkScanInternalLoop()
                .then(() => {
                config_1.log.silly('PuppetPadchatManager', `startCheckScan() checkScanInternalLoop() resolved`);
            })
                .catch(e => {
                config_1.log.warn('PuppetPadchatManager', 'startCheckScan() checkScanLoop() exception: %s', e);
                this.reset('startCheckScan() checkScanLoop() exception');
            });
            config_1.log.silly('PuppetPadchatManager', `startCheckScan() checkScanInternalLoop() set`);
        });
    }
    /**
     * Offline, then relogin
     * emit qrcode or send login request to the user.
     */
    tryAutoLogin(memorySlot) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `tryAutoLogin(%s)`, memorySlot.currentUserId);
            const currentUserId = memorySlot.currentUserId;
            if (!currentUserId) {
                config_1.log.silly('PuppetPadchatManager', 'tryAutoLogin() currentUserId not found in memorySlot');
                return false;
            }
            const deviceInfo = memorySlot.device[currentUserId];
            if (!deviceInfo) {
                config_1.log.silly('PuppetPadchatManager', 'tryAutoLogin() deviceInfo not found for userId "%s"', currentUserId);
                return false;
            }
            const token = deviceInfo.token;
            if (!token) {
                config_1.log.silly('PuppetPadchatManager', 'tryAutoLogin() token not found for userId "%s"', currentUserId);
                return false;
            }
            /**
             * PadchatRpc WXAutoLogin result:
             * {
             *  "email": "",
             *  "external": "",
             *  "long_link_server": "hklong.weixin.qq.com",
             *  "message": "\n�\u0003<e>\n<ShowType>1</ShowType>\n<Content>
             *    <![CDATA[当前帐号于02:10在iPad设备上登录。若非本人操作，你的登录密码可能已经泄漏，请及时改密。紧急情况可前往http://weixin110.qq.com冻结帐号。]]></Content>
             *    \n<Url><![CDATA[]]></Url>\n<DispSec>30</DispSec>\n<Title><![CDATA[]]></Title>\n<Action>4</Action>\n<DelayConnSec>0</DelayConnSec>
             *    \n<Countdown>0</Countdown>\n<Ok><![CDATA[]]></Ok>\n<Cancel><![CDATA[]]></Cancel>\n</e>\n",
             *  "nick_name": "",
             *  "phone_number": "",
             *  "qq": 0,
             *  "short_link_server": "hkshort.weixin.qq.com:80",
             *  "status": -2023,
             *  "uin": 1928023446,
             *  "user_name": "wxid_5zj4i5htp9ih22"
             * }
             */
            /**
             * WXAutoLoginresult: {
             *   "email": "",
             *   "external": "",
             *   "long_link_server": "",
             *   "message": "\n�\u0002<e>\n<ShowType>1</ShowType>\n<Content><![CDATA[你已退出微信]]></Content>\n
             *     <Url><![CDATA[]]></Url>\n<DispSec>30</DispSec>\n<Title><![CDATA[]]></Title>\n<Action>4</Action>\n
             *     <DelayConnSec>0</DelayConnSec>\n<Countdown>0</Countdown>\n<Ok><![CDATA[]]></Ok>\n<Cancel><![CDATA[]]></Cancel>\n</e>\n",
             *   "nick_name": "",
             *   "phone_number": "",
             *   "qq": 0,
             *   "short_link_server": "",
             *   "status": -2023,
             *   "uin": 4763975,
             *   "user_name": "lizhuohuan"
             * }
             */
            /**
             * 1. Auto Login data invalid, emit QrCode for scan
             */
            const autoLoginResult = yield this.WXAutoLogin(token);
            if (!autoLoginResult) {
                /**
                 * 1.1 Delete token for prevent future useless auto login retry
                 */
                delete deviceInfo.token;
                yield this.options.memory.set(MEMORY_SLOT_NAME, memorySlot);
                yield this.options.memory.save();
                yield this.emitLoginQrcode();
                return false;
            }
            if (autoLoginResult.status === -2023) {
                delete deviceInfo.token;
                yield this.options.memory.set(MEMORY_SLOT_NAME, memorySlot);
                yield this.options.memory.save();
                this.emit('reset');
                return false;
            }
            /**
             * 2 Auto Login Success
             */
            if (autoLoginResult.status === 0) {
                yield this.onLogin(autoLoginResult.user_name);
                return true;
            }
            /**
             * 3. Auto Login data valid, but need user to confirm.
             *  Send Login Request to User to be confirm(the same as the user had scaned the QrCode)
             *  with a valid Login Request, wait user to confirm on the phone.
             */
            const loginRequestResult = yield this.WXLoginRequest(token);
            if (loginRequestResult && loginRequestResult.status === 0) {
                return false;
            }
            /**
             * 4 Send Login Request to user fail, emit QrCode for scan.
             */
            yield this.emitLoginQrcode();
            /**
             * 5 Delete token for prevent future useless auto login retry
             */
            delete deviceInfo.token;
            yield this.options.memory.set(MEMORY_SLOT_NAME, memorySlot);
            yield this.options.memory.save();
            return false;
        });
    }
    emitLoginQrcode() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `emitLoginQrCode()`);
            if (this.loginScanQrcode) {
                throw new Error('qrcode exist');
            }
            const result = yield this.WXGetQRCode();
            if (!result || !result.qr_code) {
                config_1.log.verbose('PuppetPadchatManager', `emitLoginQrCode() result not found. Call WXInitialize() and try again ...`);
                yield this.WXInitialize();
                // wait 1 second and try again
                yield new Promise(r => setTimeout(r, 1000));
                yield this.emitLoginQrcode();
                return;
            }
            const fileBox = file_box_1.FileBox.fromBase64(result.qr_code, 'qrcode.jpg');
            const qrcodeDecoded = yield pure_function_helpers_1.fileBoxToQrcode(fileBox);
            this.loginScanQrcode = qrcodeDecoded;
            this.loginScanStatus = padchat_rpc_type_1.WXCheckQRCodeStatus.WaitScan;
            this.emit('scan', this.loginScanQrcode, this.loginScanStatus);
        });
    }
    refreshMemorySlotData(memorySlot, userId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `refresh62Data(%s, %s)`, userId);
            /**
             * must do a HeatBeat before WXGenerateWxData()
             */
            yield this.WXHeartBeat();
            /**
             * 1. Empty memorySlot, init & return it
             */
            if (!memorySlot.currentUserId) {
                config_1.log.silly('PuppetPadchatManager', 'refresh62Data() memorySlot is empty, init & return it');
                const data = yield this.WXGenerateWxDat();
                const token = yield this.WXGetLoginToken();
                memorySlot.currentUserId = userId;
                memorySlot.device[userId] = {
                    data,
                    token,
                };
                return memorySlot;
            }
            /**
             * 2. User account not changed between this and the last login session
             */
            if (memorySlot.currentUserId === userId) {
                config_1.log.silly('PuppetPadchatManager', 'refresh62Data() userId did not change since last login, keep the data as the same');
                // Update Token
                memorySlot.device[userId].token = yield this.WXGetLoginToken();
                return memorySlot;
            }
            /**
             * 3. Current user is a user that had used this memorySlot, use the old data for it.
             */
            if (userId in memorySlot.device) {
                config_1.log.silly('PuppetPadchatManager', 'refresh62Data() current userId has existing device info, set %s as currentUserId and use old data for it', userId);
                memorySlot.currentUserId = userId;
                memorySlot.device[userId] = Object.assign({}, memorySlot.device[userId], { token: yield this.WXGetLoginToken() });
                return memorySlot;
            }
            /**
             * 4. New user login, generate 62data for it
             */
            // Build a new code block to make tslint happy: no-shadow-variable
            {
                config_1.log.verbose('PuppetPadchatManager', 'refresh62Data() user switch detected: from "%s" to "%s"', memorySlot.currentUserId, userId);
                const data = yield this.WXGenerateWxDat();
                const token = yield this.WXGetLoginToken();
                memorySlot.currentUserId = userId;
                memorySlot.device[userId] = {
                    data,
                    token,
                };
                return memorySlot;
            }
        });
    }
    tryLoad62Data() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `tryLoad62Data()`);
            const deviceUserId = this.memorySlot.currentUserId;
            const deviceInfoDict = this.memorySlot.device;
            if (deviceUserId
                && deviceInfoDict
                && deviceUserId in deviceInfoDict) {
                const deviceInfo = deviceInfoDict[deviceUserId];
                if (!deviceInfo) {
                    throw new Error('deviceInfo not found');
                }
                config_1.log.silly('PuppetPadchatManager', `tryLoad62Data() 62 data found: "%s"`, deviceInfo.data);
                yield this.WXLoadWxDat(deviceInfo.data);
            }
            else {
                config_1.log.silly('PuppetPadchatManager', `tryLoad62Data() 62 data not found`);
            }
        });
    }
    getContactIdList() {
        config_1.log.verbose('PuppetPadchatManager', 'getContactIdList()');
        if (!this.cacheContactRawPayload) {
            throw new Error('cache not inited');
        }
        const contactIdList = [...this.cacheContactRawPayload.keys()];
        config_1.log.silly('PuppetPadchatManager', 'getContactIdList() = %d', contactIdList.length);
        return contactIdList;
    }
    getRoomIdList() {
        config_1.log.verbose('PuppetPadchatManager', 'getRoomIdList()');
        if (!this.cacheRoomRawPayload) {
            throw new Error('cache not inited');
        }
        const roomIdList = [...this.cacheRoomRawPayload.keys()];
        config_1.log.verbose('PuppetPadchatManager', 'getRoomIdList()=%d', roomIdList.length);
        return roomIdList;
    }
    roomMemberRawPayloadDirty(roomId) {
        config_1.log.verbose('PuppetPadchatManager', 'roomMemberRawPayloadDirty(%d)', roomId);
        if (!this.cacheRoomMemberRawPayload) {
            throw new Error('cache not inited');
        }
        this.cacheRoomMemberRawPayload.delete(roomId);
    }
    getRoomMemberIdList(roomId, dirty = false) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', 'getRoomMemberIdList(%d)', roomId);
            if (!this.cacheRoomMemberRawPayload) {
                throw new Error('cache not inited');
            }
            if (dirty) {
                this.roomMemberRawPayloadDirty(roomId);
            }
            const memberRawPayloadDict = this.cacheRoomMemberRawPayload.get(roomId)
                || (yield this.syncRoomMember(roomId));
            if (!memberRawPayloadDict) {
                // or return [] ?
                throw new Error('roomId not found: ' + roomId);
            }
            const memberIdList = Object.keys(memberRawPayloadDict);
            // console.log('memberRawPayloadDict:', memberRawPayloadDict)
            config_1.log.verbose('PuppetPadchatManager', 'getRoomMemberIdList(%d) length=%d', roomId, memberIdList.length);
            return memberIdList;
        });
    }
    roomRawPayloadDirty(roomId) {
        config_1.log.verbose('PuppetPadchatManager', 'roomRawPayloadDirty(%d)', roomId);
        if (!this.cacheRoomRawPayload) {
            throw new Error('cache not inited');
        }
        this.cacheRoomRawPayload.delete(roomId);
    }
    roomMemberRawPayload(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', 'roomMemberRawPayload(%s)', roomId);
            if (!this.cacheRoomMemberRawPayload) {
                throw new Error('cache not inited');
            }
            const memberRawPayloadDict = this.cacheRoomMemberRawPayload.get(roomId)
                || (yield this.syncRoomMember(roomId));
            if (!memberRawPayloadDict) {
                throw new Error('roomId not found: ' + roomId);
            }
            return memberRawPayloadDict;
        });
    }
    syncRoomMember(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PuppetPadchatManager', 'syncRoomMember(%s)', roomId);
            const memberListPayload = yield this.WXGetChatRoomMember(roomId);
            if (!memberListPayload || !('user_name' in memberListPayload)) { // check user_name too becasue the server might return {}
                /**
                 * Room Id not exist
                 * See: https://github.com/lijiarui/wechaty-puppet-padchat/issues/64#issuecomment-397319016
                 */
                this.roomMemberRawPayloadDirty(roomId);
                this.roomRawPayloadDirty(roomId);
                return {};
            }
            config_1.log.silly('PuppetPadchatManager', 'syncRoomMember(%s) total %d members', roomId, Object.keys(memberListPayload).length);
            const memberDict = {};
            for (const memberPayload of memberListPayload.member) {
                const contactId = memberPayload.user_name;
                memberDict[contactId] = memberPayload;
                const contact = {
                    big_head: memberPayload.big_head,
                    city: '',
                    country: '',
                    intro: '',
                    label: '',
                    nick_name: memberPayload.nick_name,
                    provincia: '',
                    py_initial: '',
                    remark: '',
                    remark_py_initial: '',
                    remark_quan_pin: '',
                    sex: wechaty_puppet_1.ContactGender.Unknown,
                    signature: '',
                    small_head: memberPayload.small_head,
                    status: padchat_schemas_1.PadchatContactRoomStatus.Sync,
                    stranger: '',
                    ticket: '',
                    user_name: memberPayload.user_name,
                    message: '',
                    quan_pin: '',
                };
                if (!this.cacheContactRawPayload) {
                    throw new Error('cache not inited');
                }
                if (!this.cacheContactRawPayload.has(contactId)) {
                    config_1.log.silly('PadchatManager', `syncRoomMember() adding room member ${contactId} into contact cache`);
                    this.cacheContactRawPayload.set(contactId, contact);
                }
            }
            if (!this.cacheRoomMemberRawPayload) {
                throw new Error('cache not inited');
            }
            const oldMemberDict = this.cacheRoomMemberRawPayload.get(roomId);
            const newMemberDict = Object.assign({}, oldMemberDict, memberDict);
            this.cacheRoomMemberRawPayload.set(roomId, newMemberDict);
            return newMemberDict;
        });
    }
    syncContactsAndRooms() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', `syncContactsAndRooms()`);
            let cont = true;
            while (cont && this.state.on() && this.userId) {
                config_1.log.silly('PuppetPadchatManager', `syncContactsAndRooms() while() syncing WXSyncContact ...`);
                yield new Promise(r => setTimeout(r, 3 * 1000));
                const syncContactList = yield this.WXSyncContact();
                if (!Array.isArray(syncContactList) || syncContactList.length <= 0) {
                    // {"apiName":"WXSyncContact","data":"%7B%22message%22%3A%22%E4%B8%8D%E6%94%AF%E6%8C%81%E7%9A%84%E8%AF%B7%E6%B1%82%22%2C%22status%22%3A-19%7D"}
                    // data => { message: '不支持的请求', status: -19 }
                    config_1.log.warn('PuppetPadchatManager', 'syncContactsAndRooms() cannot get array result: %s', JSON.stringify(syncContactList));
                    cont = false;
                    return;
                }
                if (!this.cacheContactRawPayload
                    || !this.cacheRoomRawPayload) {
                    throw new Error('no cache');
                }
                config_1.log.silly('PuppetPadchatManager', 'syncContactsAndRooms() syncing %d out of Contact(%d) & Room(%d) ...', syncContactList.length, this.cacheContactRawPayload.size, this.cacheRoomRawPayload.size);
                for (const syncContact of syncContactList) {
                    if (syncContact.continue !== padchat_schemas_1.PadchatContinue.Go) {
                        config_1.log.verbose('PuppetPadchatManager', 'syncContactsAndRooms() sync contact done! But rooms might be still syncing!');
                        cont = false;
                        this.contactListSynced = true;
                        break;
                    }
                    if (syncContact.msg_type === padchat_schemas_1.PadchatContactMsgType.Contact) {
                        if (pure_function_helpers_1.isRoomId(syncContact.user_name)) {
                            /**
                             * Room
                             */
                            config_1.log.silly('PuppetPadchatManager', 'syncContactsAndRooms() updating Room %s(%s)', syncContact.nick_name, syncContact.user_name);
                            const roomId = syncContact.user_name;
                            const roomPayload = syncContact;
                            this.cacheRoomRawPayload.set(roomId, roomPayload);
                            /**
                             * Use delay queue executor to sync room:
                             *  add syncRoomMember task to the queue
                             */
                            this.roomNeedsToBeSync++;
                            this.delayQueueExecutor.execute(() => {
                                this.roomNeedsToBeSync--;
                                return this.syncRoomMember(roomId);
                            }, `syncRoomMember(${roomId})`).catch((e) => {
                                config_1.log.error('PuppetPadchatManager', 'syncContactsAndRooms() failed to process %s: %s', roomId, e);
                            });
                            config_1.log.silly('PuppetPadchatManager', 'syncContactsAndRooms() added sync room(%s) task to delayQueueExecutor', roomId);
                        }
                        else if (pure_function_helpers_1.isContactId(syncContact.user_name)) {
                            /**
                             * Contact
                             */
                            config_1.log.silly('PuppetPadchatManager', 'syncContactsAndRooms() updating Contact %s(%s)', syncContact.nick_name, syncContact.user_name);
                            const contactPayload = syncContact;
                            const contactId = contactPayload.user_name;
                            this.cacheContactRawPayload.set(contactId, contactPayload);
                        }
                        else {
                            throw new Error('id is neither room nor contact');
                        }
                    }
                    else {
                        // {"continue":1,"msg_type":2048,"status":1,"uin":4763975}
                        if (syncContact.continue === padchat_schemas_1.PadchatContinue.Go
                            && syncContact.msg_type === padchat_schemas_1.PadchatContactMsgType.N11_2048
                            && typeof syncContact.uin !== 'undefined') {
                            // HeartBeat??? discard it in silent
                        }
                        else {
                            config_1.log.silly('PuppetPadchatManager', `syncContactsAndRooms() skip for syncContact.msg_type=%s(%s) %s`, syncContact.msg_type && padchat_schemas_1.PadchatContactMsgType[syncContact.msg_type], syncContact.msg_type, JSON.stringify(syncContact));
                        }
                    }
                }
            }
        });
    }
    contactRawPayloadDirty(contactId) {
        config_1.log.verbose('PuppetPadchatManager', 'contactRawPayloadDirty(%d)', contactId);
        if (!this.cacheContactRawPayload) {
            throw new Error('cache not inited');
        }
        this.cacheContactRawPayload.delete(contactId);
    }
    contactRawPayload(contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PuppetPadchatManager', 'contactRawPayload(%s)', contactId);
            const rawPayload = yield config_1.retry((retryException, attempt) => __awaiter(this, void 0, void 0, function* () {
                config_1.log.silly('PuppetPadchatManager', 'contactRawPayload(%s) retry() attempt=%d', contactId, attempt);
                if (!this.cacheContactRawPayload) {
                    throw new Error('no cache');
                }
                if (this.cacheContactRawPayload.has(contactId)) {
                    return this.cacheContactRawPayload.get(contactId);
                }
                const tryRawPayload = yield this.WXGetContactPayload(contactId);
                // check user_name too becasue the server might return {}
                // See issue #1358 https://github.com/Chatie/wechaty/issues/1358
                if (tryRawPayload && tryRawPayload.user_name) {
                    this.cacheContactRawPayload.set(contactId, tryRawPayload);
                    return tryRawPayload;
                }
                else if (tryRawPayload) {
                    // If the payload is valid but we don't have user_name inside it,
                    // consider this payload as invalid one and do not retry
                    // Correct me if I am wrong here
                    return null;
                }
                return retryException(new Error('tryRawPayload empty'));
            }));
            if (!rawPayload) {
                throw new Error('no raw payload');
            }
            return rawPayload;
        });
    }
    roomRawPayload(id) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', 'roomRawPayload(%s)', id);
            const rawPayload = yield config_1.retry((retryException, attempt) => __awaiter(this, void 0, void 0, function* () {
                config_1.log.silly('PuppetPadchatManager', 'roomRawPayload(%s) retry() attempt=%d', id, attempt);
                if (!this.cacheRoomRawPayload) {
                    throw new Error('no cache');
                }
                if (this.cacheRoomRawPayload.has(id)) {
                    return this.cacheRoomRawPayload.get(id);
                }
                const tryRawPayload = yield this.WXGetRoomPayload(id);
                // check user_name too becasue the server might return {}
                // See issue #1358 https://github.com/Chatie/wechaty/issues/1358
                if (tryRawPayload /* && tryRawPayload.user_name */) {
                    this.cacheRoomRawPayload.set(id, tryRawPayload);
                    return tryRawPayload;
                }
                return retryException(new Error('tryRawPayload empty'));
            }));
            if (!rawPayload) {
                throw new Error('no raw payload');
            }
            return rawPayload;
        });
    }
    roomInvitationRawPayload(roomInvitationId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', 'roomInvitationRawPayload(%s)', roomInvitationId);
            if (!this.cacheRoomInvitationRawPayload) {
                throw new Error('no cache');
            }
            const payload = yield this.cacheRoomInvitationRawPayload.get(roomInvitationId);
            if (payload) {
                return payload;
            }
            else {
                throw new Error(`can not get invitation with invitation id: ${roomInvitationId}`);
            }
        });
    }
    roomInvitationRawPayloadDirty(roomInvitationId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', 'roomInvitationRawPayloadDirty(%s)', roomInvitationId);
            if (!this.cacheRoomInvitationRawPayload) {
                throw new Error('no cache');
            }
            yield this.cacheRoomInvitationRawPayload.delete(roomInvitationId);
        });
    }
    saveRoomInvitationRawPayload(roomInvitation) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchatManager', 'saveRoomInvitationRawPayload(%s)', JSON.stringify(roomInvitation));
            const { msgId, roomName, url, fromUser, timestamp } = roomInvitation;
            if (!this.cacheRoomInvitationRawPayload) {
                throw new Error('no cache');
            }
            this.cacheRoomInvitationRawPayload.set(msgId, {
                fromUser,
                id: msgId,
                roomName,
                timestamp,
                url,
            });
        });
    }
    ding(data) {
        config_1.log.verbose('PuppetPadchatManager', 'ding(%s)', data || '');
        // TODO: healthy check
        this.emit('dong');
        return;
    }
    updateSelfName(newName) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.userId) {
                throw Error('Can not update user self name since no user id exist. Probably user not logged in yet');
            }
            const self = yield this.contactRawPayload(this.userId);
            const { signature, sex, country, provincia, city } = self;
            yield this.WXSetUserInfo(newName, signature, sex.toString(), country, provincia, city);
            this.contactRawPayloadDirty(this.userId);
        });
    }
    updateSelfSignature(signature) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.userId) {
                throw Error('Can not update user self signature since no user id exist. Probably user not logged in yet');
            }
            const self = yield this.contactRawPayload(this.userId);
            const { nick_name, sex, country, provincia, city } = self;
            yield this.WXSetUserInfo(nick_name, signature, sex.toString(), country, provincia, city);
            this.contactRawPayloadDirty(this.userId);
        });
    }
}
exports.PadchatManager = PadchatManager;
exports.default = PadchatManager;
//# sourceMappingURL=padchat-manager.js.map