/// <reference types="node" />
import { EventEmitter } from 'events';
import { PadchatContactPayload, PadchatMessagePayload, PadchatPayload, PadchatRequestTokenPayload, PadchatRoomMemberListPayload, PadchatRoomPayload } from './padchat-schemas';
import { InitType, StandardType, WXAutoLoginType, WXCheckQRCodePayload, WXGetQRCodeType, WXHeartBeatType, WXInitializeType, WXLoadWxDatType, WXLoginRequestType, WXLogoutType, WXQRCodeLoginType, WXSearchContactType, WXSearchContactTypeStatus, WXSendMsgType } from './padchat-rpc.type';
export declare const DISCONNECTED = "DISCONNECTED";
export declare const CONNECTING = "CONNECTING";
export declare const CONNECTED = "CONNECTED";
export interface ConnectionStatus {
    status: string;
    reconnectLeft: number;
    interval: number;
}
export interface PendingAPICall {
    apiName: string;
    params: Array<number | string>;
    timestamp: number;
    resolve: (value?: any | PromiseLike<any>) => void;
    reject: (reason?: any) => void;
}
export declare class PadchatRpc extends EventEmitter {
    protected endpoint: string;
    protected token: string;
    private socket?;
    private readonly jsonRpc;
    private throttleQueue?;
    private debounceQueue?;
    private reconnectThrottleQueue?;
    private resetThrottleQueue?;
    private throttleSubscription?;
    private debounceSubscription?;
    private reconnectThrottleSubscription?;
    private resetThrottleSubscription?;
    private pendingApiCalls;
    protected connectionStatus: ConnectionStatus;
    protected userId?: string;
    constructor(endpoint: string, token: string);
    start(): Promise<void>;
    protected reconnect(): Promise<void>;
    protected updateConnectionStatus(status: string): void;
    protected cleanConnection(): void;
    private executeBufferedApiCalls;
    protected initJsonRpc(): Promise<void>;
    protected initWebSocket(): Promise<void>;
    private initHeartbeat;
    protected reset(reason?: string): void;
    stop(): void;
    private startQueues;
    private stopQueues;
    private rpcCall;
    protected onSocket(payload: PadchatPayload): void;
    protected onSocketPadchat(padchatPayload: PadchatPayload): void;
    protected onSocketTencent(messagePayloadList: PadchatMessagePayload[]): void;
    private replayTextMsg;
    private replayImageMsg;
    private replayAppMsg;
    private generateBaseMsg;
    /**
     * Init with WebSocket Server
     */
    protected init(): Promise<InitType>;
    /**
     * Get WX block memory
     */
    protected WXInitialize(): Promise<WXInitializeType>;
    protected WXGetQRCode(): Promise<WXGetQRCodeType>;
    WXCheckQRCode(): Promise<WXCheckQRCodePayload>;
    WXHeartBeat(): Promise<WXHeartBeatType>;
    /**
     * Load all Contact and Room
     * see issue https://github.com/lijiarui/wechaty-puppet-padchat/issues/39
     * @returns {Promise<(PadchatRoomPayload | PadchatContactPayload)[]>}
     */
    WXSyncContact(): Promise<Array<PadchatRoomPayload | PadchatContactPayload>>;
    /**
     * Generate 62 data
     *
     * 1. Call multiple times in the same session, will return the same data
     * 2. Call multiple times between sessions with the same token, will return the same data
     */
    WXGenerateWxDat(): Promise<string>;
    /**
     * Load 62 data
     * @param {string} wxData     autoData.wxData
     */
    WXLoadWxDat(wxData: string): Promise<WXLoadWxDatType>;
    WXGetLoginToken(): Promise<string>;
    /**
     * Login with token automatically
     * @param {string} token    autoData.token
     * @returns {string} user_name | ''
     */
    WXAutoLogin(token: string): Promise<undefined | WXAutoLoginType>;
    /**
     * Login with QRcode
     * @param {string} token    autoData.token
     */
    WXLoginRequest(token: string): Promise<undefined | WXLoginRequestType>;
    /**
     * Send Text Message
     * @param {string} to       user_name
     * @param {string} content  text
     */
    WXSendMsg(to: string, content: string, at?: string[]): Promise<WXSendMsgType>;
    /**
     * Send Image Message
     * @param {string} to     user_name
     * @param {string} data   image_data
     */
    WXSendImage(to: string, data: string): Promise<void>;
    /**
     * Get contact by contact id
     * @param {any} id        user_name
     */
    WXGetContact(id: string): Promise<any>;
    /**
     * Get contact by contact id
     * @param {any} id        user_name
     */
    WXGetContactPayload(id: string): Promise<PadchatContactPayload>;
    /**
     * Get contact by contact id
     * @param {any} id        user_name
     */
    WXGetRoomPayload(id: string): Promise<PadchatRoomPayload>;
    /**
     * Get all member of a room by room id
     * @param {any} roomId        chatroom_id
     */
    WXGetChatRoomMember(roomId: string): Promise<null | PadchatRoomMemberListPayload>;
    /**
     * Login successfully by qrcode
     * @param {any} username        user_name
     * @param {any} password  password
     */
    WXQRCodeLogin(username: string, password: string): Promise<WXQRCodeLoginType>;
    WXSetUserRemark(id: string, remark: string): Promise<StandardType>;
    WXDeleteChatRoomMember(roomId: string, contactId: string): Promise<StandardType>;
    WXAddChatRoomMember(roomId: string, contactId: string): Promise<number>;
    /**
     * When member more than 40, use WXInviteChatRoomMember
     * When member less than 40, use WXAddChatRoomMember
     *
     * @param {string} roomId
     * @param {string} contactId
     * @returns {Promise<number>}
     */
    WXInviteChatRoomMember(roomId: string, contactId: string): Promise<number>;
    WXSetChatroomName(roomId: string, topic: string): Promise<StandardType>;
    WXQuitChatRoom(roomId: string): Promise<StandardType>;
    WXAddUser(strangerV1: string, strangerV2: string, type: WXSearchContactTypeStatus, verify: string): Promise<any>;
    WXAcceptUser(stranger: string, ticket: string): Promise<any>;
    WXLogout(): Promise<WXLogoutType>;
    WXSendMoments(text: string): Promise<any>;
    WXSyncMessage(): Promise<any>;
    WXSearchContact(id: string): Promise<WXSearchContactType>;
    WXSayHello(stranger: string, text: string): Promise<any>;
    WXDeleteUser(id: string): Promise<any>;
    WXCreateChatRoom(userList: string[]): Promise<string>;
    WXReceiveRedPacket(redPacket: string): Promise<any>;
    WXQueryRedPacket(redPacket: string): Promise<any>;
    WXOpenRedPacket(redPacket: string): Promise<any>;
    WXSetHeadImage(image: string): Promise<any>;
    WXSetUserInfo(nickeName: string, signature: string, sex: string, country: string, provincia: string, city: string): Promise<any>;
    WXGetPeopleNearby(longitude: string, latitude: string): Promise<any>;
    WXGetSubscriptionInfo(id: string): Promise<any>;
    WXSubscriptionCommand(id: string, orderId: string, orderKey: string): Promise<any>;
    WXGetRequestToken(id: string, url: string): Promise<PadchatRequestTokenPayload>;
    WXRequestUrl(url: string, key: string, uin: string): Promise<any>;
    WXSetWeChatID(id: string): Promise<any>;
    WXTransferQuery(transfer: string): Promise<any>;
    WXTransferOperation(transfer: string): Promise<any>;
    WXGetMsgImage(msg: string): Promise<any>;
    WXGetMsgVideo(msg: string): Promise<any>;
    WXGetMsgVoice(msg: string): Promise<any>;
    WXWebSearch(id: string): Promise<any>;
    WXShareCard(user: string, id: string, caption: string): Promise<any>;
    WXSyncReset(): Promise<any>;
    WXQRCodeDecode(path: string): Promise<any>;
    WXSnsUpload(image: string): Promise<any>;
    WXSnsObjectDetail(id: string): Promise<any>;
    WXSnsObjectOp(id: string, type: string, comment: string, commentType: string): Promise<any>;
    WXSnsComment(user: string, id: string, content: string, replyId: string): Promise<any>;
    WXSnsUserPage(user: string, id: string): Promise<any>;
    WXSnsTimeline(id: string): Promise<any>;
    WXSendAppMsg(user: string, content: string): Promise<any>;
    WXFavSync(key: string): Promise<any>;
    WXFavAddItem(favObject: string): Promise<any>;
    WXFavGetItem(id: string): Promise<any>;
    WXFavDeleteItem(id: string): Promise<any>;
    WXGetContactLabelList(): Promise<any>;
    WXAddContactLabel(label: string): Promise<any>;
    WXDeleteContactLabel(id: string): Promise<any>;
    WXSetContactLabel(user: string, id: string): Promise<any>;
    WXGetUserQRCode(user: string, style: number): Promise<string>;
    WXUploadAppAttach(data: string): Promise<any>;
    WXSendVoice(user: string, data: string, time: number): Promise<any>;
    WXSnsSync(key: string): Promise<any>;
    WXMassMessage(userList: string[], content: string): Promise<any>;
    WXSetChatroomAnnouncement(chatroom: string, content: string): Promise<any>;
}
