"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const json_rpc_peer_1 = __importStar(require("json-rpc-peer"));
const ws_1 = __importDefault(require("ws"));
const rx_queue_1 = require("rx-queue");
const padchat_schemas_1 = require("./padchat-schemas");
const padchat_rpc_type_1 = require("./padchat-rpc.type");
const pure_function_helpers_1 = require("./pure-function-helpers/");
const config_1 = require("./config");
const rpc_api_1 = require("./rpc-api");
let HEART_BEAT_COUNTER = 0;
let RPC_TIMEOUT_COUNTER = 0;
exports.DISCONNECTED = 'DISCONNECTED';
exports.CONNECTING = 'CONNECTING';
exports.CONNECTED = 'CONNECTED';
class PadchatRpc extends events_1.EventEmitter {
    constructor(endpoint, token) {
        super(); // for EventEmitter
        this.endpoint = endpoint;
        this.token = token;
        config_1.log.verbose('PadchatRpc', 'constructor(%s, %s)', endpoint, token);
        this.connectionStatus = {
            interval: config_1.RE_CON_INTERVAL,
            reconnectLeft: config_1.RE_CON_RETRY,
            status: exports.DISCONNECTED,
        };
        this.pendingApiCalls = [];
        this.jsonRpc = new json_rpc_peer_1.default();
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PadchatRpc', 'start()');
            yield this.initWebSocket();
            yield this.initJsonRpc();
            this.startQueues();
            yield this.init();
            yield this.WXInitialize();
        });
    }
    reconnect() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.info('PadchatRpc', 'reconnect()');
            this.cleanConnection();
            yield this.initWebSocket();
            yield this.init();
            yield this.WXInitialize();
        });
    }
    updateConnectionStatus(status) {
        let reconnectLeft;
        if (status === exports.CONNECTING) {
            reconnectLeft = this.connectionStatus.reconnectLeft - 1;
        }
        else {
            reconnectLeft = config_1.RE_CON_RETRY;
        }
        if (status === exports.CONNECTED && this.connectionStatus.status === exports.CONNECTING) {
            this.executeBufferedApiCalls();
        }
        this.connectionStatus.status = status;
        this.connectionStatus.reconnectLeft = reconnectLeft;
    }
    cleanConnection() {
        if (this.socket && this.socket.readyState === 1) {
            this.socket.removeAllListeners();
            this.socket.close();
        }
        this.socket = undefined;
    }
    executeBufferedApiCalls() {
        const timer = setTimeout(() => {
            clearTimeout(timer);
            if (this.pendingApiCalls.length > 0) {
                const apiCallsInOrder = this.pendingApiCalls.sort((a, b) => {
                    return a.timestamp - b.timestamp;
                });
                apiCallsInOrder.map((api, index) => {
                    const id = setTimeout(() => {
                        clearTimeout(id);
                        this.jsonRpc.request(api.apiName, api.params)
                            .then((res) => {
                            api.resolve(res);
                        }).catch((reason) => {
                            api.reject(reason);
                        });
                    }, config_1.POST_LOGIN_API_CALL_INTERVAL * index);
                });
                this.pendingApiCalls = [];
            }
        }, 1000);
    }
    initJsonRpc() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PadchatRpc', 'initJsonRpc()');
            if (!this.socket) {
                throw new Error('socket had not been opened yet!');
            }
            this.jsonRpc.on('error', () => {
                // TypeError: Cannot read property 'resolve' of undefined
                // https://github.com/JsCommunity/json-rpc-peer/issues/52
            });
            this.jsonRpc.on('data', (buffer) => {
                // log.silly('PadchatRpc', 'initJsonRpc() jsonRpc.on(data)')
                if (!this.socket) {
                    throw new Error('no web socket');
                }
                const text = String(buffer);
                const payload = json_rpc_peer_1.parse(text); // as JsonRpcPayloadRequest
                /**
                 * A Gateway at here:
                 *
                 *  1. Convert Payload format from JsonRpc to Padchat, then
                 *  2. Send payload to padchat server
                 *
                 */
                // const encodedParam = (payload.params as JsonRpcParamsSchemaByPositional).map(encodeURIComponent)
                const encodedParam = payload.params.map(encodeURIComponent);
                const message = {
                    apiName: payload.method,
                    msgId: payload.id,
                    param: encodedParam,
                    userId: this.token,
                };
                // log.silly('PadchatRpc', 'initJsonRpc() jsonRpc.on(data) converted to padchat payload="%s"', JSON.stringify(message))
                this.socket.send(JSON.stringify(message));
            });
        });
    }
    initWebSocket() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PadchatRpc', 'initWebSocket()');
            if (this.socket) {
                throw new Error('socket had already been opened!');
            }
            const ws = new ws_1.default(this.endpoint, { perMessageDeflate: true, maxPayload: 100 * 1024 * 1024 });
            /**
             * 1. Message
             *  1.1. Deal with payload
             */
            ws.on('message', (data) => {
                config_1.log.silly('PadchatRpc', 'initWebSocket() ws.on(message): %s', data.substr(0, 100));
                try {
                    const payload = JSON.parse(data);
                    this.onSocket(payload);
                }
                catch (e) {
                    config_1.log.warn('PadchatRpc', 'initWebSocket() ws.on(message) exception: %s', e);
                    this.emit('error', e);
                }
            });
            /**
             * 1. Message
             *  1.2. use websocket message as heartbeat source
             */
            ws.on('message', () => {
                if (!this.throttleQueue || !this.debounceQueue) {
                    config_1.log.warn('PadchatRpc', 'initWebSocket() ws.on(message) throttleQueue or debounceQueue not exist');
                    return;
                }
                if (this.connectionStatus.status === exports.CONNECTED) {
                    this.debounceQueue.next('ws.on(message)');
                    this.throttleQueue.next('ws.on(message)');
                }
            });
            /**
             * 2. Error
             */
            ws.on('error', e => {
                if (e.message.indexOf('ECONNREFUSED') !== -1) {
                    // Can not connect to remote server, if this is triggered when puppet-padchat is connected,
                    // an close event must be emitted, so ignore this error
                    // If this is triggered when puppet-padchat is trying to reconnect, also ignore this error
                }
                else {
                    config_1.log.verbose('PadchatRpc', 'initWebSocket() ws.on(error) %s', e);
                    this.emit('error', e);
                }
            });
            /**
             * 3. Close
             */
            ws.on('close', (code, reason) => {
                config_1.log.warn('PadchatRpc', 'initWebSocket() ws.on(close) code: %s, reason: %s', code, reason);
                if (!this.reconnectThrottleQueue) {
                    config_1.log.warn('PadchatRpc', 'initWebSocket() ws.on(close) reconnectThrottleQueue not exist');
                    return;
                }
                if (this.connectionStatus.status === exports.CONNECTED) {
                    this.reconnectThrottleQueue.next('ws.on(close, ' + code + ')');
                }
            });
            /**
             * 4. Pong
             */
            ws.on('pong', data => {
                config_1.log.silly('PadchatRpc', 'initWebSocket() ws.on(pong)');
                if (!this.throttleQueue || !this.debounceQueue) {
                    config_1.log.warn('PadchatRpc', 'initWebSocket() ws.on(pong) throttleQueue or debounceQueue not exist');
                    return;
                }
                if (this.connectionStatus.status === exports.CONNECTED) {
                    this.throttleQueue.next('pong: ' + data.toString());
                    this.debounceQueue.next('pong: ' + data.toString());
                }
            });
            /**
             * 5. Wait the WebSocket to be connected
             */
            yield new Promise((resolve, reject) => {
                ws.once('open', () => {
                    config_1.log.silly('PadchatRpc', 'initWebSocket() Promise() ws.on(open)');
                    return resolve();
                });
                ws.once('error', (e) => {
                    config_1.log.silly('PadchatRpc', 'initWebSocket() Promise() ws.on(error) %s', e);
                    return reject();
                });
                ws.once('close', () => {
                    config_1.log.silly('PadchatRpc', 'initWebSocket() Promise() ws.on(close)');
                    return reject();
                });
            });
            /**
             * 6. Set socket to the new WebSocket instance
             */
            this.socket = ws;
        });
    }
    initHeartbeat() {
        config_1.log.verbose('PadchatRpc', 'initHeartbeat()');
        if (!this.throttleQueue || !this.debounceQueue) {
            config_1.log.warn('PadchatRpc', 'initHeartbeat() throttleQueue or debounceQueue not exist');
            return;
        }
        if (this.throttleSubscription || this.debounceSubscription) {
            throw new Error('subscription exist when initHeartbeat');
        }
        this.throttleSubscription = this.throttleQueue.subscribe(e => {
            /**
             * This block will only be run once in a period,
             *  no matter than how many message the queue received.
             */
            config_1.log.silly('PadchatRpc', 'initHeartbeat() throttleQueue.subscribe(%s)', e);
            this.emit('heartbeat', e);
        });
        this.debounceSubscription = this.debounceQueue.subscribe(e => {
            /**
             * This block will be run when:
             *  the queue did not receive any message after a period.
             */
            if (this.connectionStatus.status !== exports.CONNECTED) {
                return;
            }
            const { socket, reconnectThrottleQueue } = this;
            config_1.log.silly('PadchatRpc', 'initHeartbeat() debounceQueue.subscribe(%s)', e);
            if (!socket) {
                throw new Error('no socket');
            }
            if (!reconnectThrottleQueue) {
                config_1.log.warn('PadchatRpc', 'initHeartbeat() debounceQueue.subscribe(%s) reconnectThrottleQueue not exist', e);
                return;
            }
            // tslint:disable-next-line:no-floating-promises
            this.WXHeartBeat().then(() => {
                RPC_TIMEOUT_COUNTER = 0;
            }).catch(reason => {
                if (reason === 'timeout' || reason === 'parsed-data-not-array') {
                    config_1.log.verbose('PadchatRpc', 'initHeartbeat() debounceQueue.subscribe(s%) WXHeartBeat %s', e, reason);
                    RPC_TIMEOUT_COUNTER++;
                    if (RPC_TIMEOUT_COUNTER >= config_1.MAX_HEARTBEAT_TIMEOUT) {
                        RPC_TIMEOUT_COUNTER = 0;
                        reconnectThrottleQueue.next(`WXHeartBeat Timed out in ${config_1.CON_TIME_OUT}ms for ${config_1.MAX_HEARTBEAT_TIMEOUT} times. Possible disconnected from Wechat. So trigger reconnect.`);
                    }
                }
                else {
                    config_1.log.verbose('PadchatRpc', 'initHeartbeat() debounceQueue.subscribe(%s) error happened: %s', e, reason);
                }
            }).finally(() => {
                // expect the server will response a 'pong' message
                socket.ping(`#${HEART_BEAT_COUNTER++} from debounceQueue`);
            });
        });
    }
    reset(reason = 'unknown reason') {
        config_1.log.verbose('PadchatRpc', 'reset(%s)', reason);
        this.emit('reset', reason);
    }
    stop() {
        config_1.log.verbose('PadchatRpc', 'stop()');
        this.stopQueues();
        this.jsonRpc.removeAllListeners();
        // TODO: use huan's version of JsonRpcPeer, to support end at here.
        // this.jsonRpc.end()
        if (this.socket) {
            this.socket.removeAllListeners();
            this.socket.close();
            this.socket = undefined;
        }
        else {
            config_1.log.warn('PadchatRpc', 'stop() no socket');
        }
    }
    startQueues() {
        config_1.log.verbose('PadchatRpc', 'startQueues()');
        /**
         * Throttle for 10 seconds
         */
        this.throttleQueue = new rx_queue_1.ThrottleQueue(1000 * 10);
        /**
         * Debounce for 20 seconds
         */
        this.debounceQueue = new rx_queue_1.DebounceQueue(1000 * 10 * 2);
        /**
         * Throttle for 5 seconds for the `reconnect` event:
         *  we should only fire once for reconnect,
         *  but many events might be triggered
         */
        this.reconnectThrottleQueue = new rx_queue_1.ThrottleQueue(1000 * 5);
        /**
         * Throttle for 5 seconds for the `reset` event:
         *  we should only fire once for reset,
         *  but the server will send many events of 'reset'
         */
        this.resetThrottleQueue = new rx_queue_1.ThrottleQueue(1000 * 5);
        this.initHeartbeat();
        if (this.reconnectThrottleSubscription) {
            throw new Error('this.reconnectThrottleSubscription exist');
        }
        else {
            this.reconnectThrottleSubscription = this.reconnectThrottleQueue.subscribe(msg => {
                if (this.connectionStatus.status === exports.CONNECTED) {
                    this.updateConnectionStatus(exports.DISCONNECTED);
                    this.jsonRpc.failPendingRequests(exports.DISCONNECTED);
                    this.emit('reconnect', msg);
                }
            });
        }
        if (this.resetThrottleSubscription) {
            throw new Error('this.reconnectThrottleSubscription exist');
        }
        else {
            this.resetThrottleSubscription = this.resetThrottleQueue.subscribe(msg => {
                if (this.connectionStatus.status === exports.CONNECTED) {
                    this.reset(msg);
                }
            });
        }
    }
    stopQueues() {
        config_1.log.verbose('PadchatRpc', 'stopQueues()');
        if (this.throttleSubscription
            && this.debounceSubscription
            && this.reconnectThrottleSubscription
            && this.resetThrottleSubscription) {
            // Clean external subscriptions
            this.debounceSubscription.unsubscribe();
            this.reconnectThrottleSubscription.unsubscribe();
            this.throttleSubscription.unsubscribe();
            this.resetThrottleSubscription.unsubscribe();
            this.debounceSubscription = undefined;
            this.reconnectThrottleSubscription = undefined;
            this.throttleSubscription = undefined;
            this.resetThrottleSubscription = undefined;
        }
        if (this.debounceQueue
            && this.reconnectThrottleQueue
            && this.resetThrottleQueue
            && this.throttleQueue) {
            /**
             * Queues clean internal subscriptions
             */
            this.debounceQueue.unsubscribe();
            this.reconnectThrottleQueue.unsubscribe();
            this.throttleQueue.unsubscribe();
            this.resetThrottleQueue.unsubscribe();
            this.debounceQueue = undefined;
            this.reconnectThrottleQueue = undefined;
            this.throttleQueue = undefined;
            this.resetThrottleQueue = undefined;
        }
        else {
            config_1.log.warn('PadchatRpc', 'stop() subscript not exist');
        }
    }
    rpcCall(apiName, ...params) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PadchatRpc', 'rpcCall(%s, %s)', apiName, JSON.stringify(params).substr(0, 500));
            const timestamp = new Date().getTime();
            // If get contact for self, this api call belongs to pre login api
            if (apiName === 'WXGetContact' && params.length === 1 && params[0] === this.userId) {
                return this.jsonRpc.request(apiName, params);
            }
            // If post login apis calls failed, we will store the api call information
            // and make the call after the connection get restored
            if (rpc_api_1.POST_LOGIN_API.indexOf(apiName) !== -1) {
                if (this.connectionStatus.status === exports.CONNECTED) {
                    return new Promise((resolve, reject) => {
                        this.jsonRpc.request(apiName, params)
                            .then((res) => {
                            resolve(res);
                        })
                            .catch((reason) => {
                            if (reason === exports.DISCONNECTED) {
                                config_1.log.verbose('PadchatRpc', 'rpcCall(%s) interrupted by connection broken, scheduled this api call. This call will be made again when connection restored', apiName);
                                this.pendingApiCalls.push({
                                    apiName,
                                    params,
                                    reject,
                                    resolve,
                                    timestamp,
                                });
                            }
                            else {
                                reject(reason);
                            }
                        });
                    });
                }
                else {
                    config_1.log.verbose('PadchatRpc', 'puppet-padchat is disconnected, rpcCall(%s) is scheduled. This call will be made again when connection restored', apiName);
                    return new Promise((resolve, reject) => {
                        this.pendingApiCalls.push({
                            apiName,
                            params,
                            reject,
                            resolve,
                            timestamp,
                        });
                    });
                }
            }
            else if (rpc_api_1.PRE_LOGIN_API.indexOf(apiName) !== -1) {
                config_1.log.silly('PadchatRpc', 'pre login rpcCall(%s, %s)', apiName, JSON.stringify(params).substr(0, 200));
                return this.jsonRpc.request(apiName, params);
            }
            else {
                config_1.log.error('PadchatRpc', 'unexpected api call: %s', apiName);
            }
        });
    }
    onSocket(payload) {
        // log.silly('PadchatRpc', 'onSocket(payload.length=%d)',
        //                           JSON.stringify(payload).length,
        //             )
        // console.log('server payload:', payload)
        if (payload.type === padchat_schemas_1.PadchatPayloadType.Logout) {
            // {"type":-1,"msg":"掉线了"}
            config_1.log.verbose('PadchatRpc', 'onSocket(payload.type=%s) logout, payload=%s(%s)', padchat_schemas_1.PadchatPayloadType[payload.type], payload.type, JSON.stringify(payload));
            // When receive this message, the bot should be logged out from the mobile phone
            // So do a full reset here
            if (this.resetThrottleQueue && this.connectionStatus.status === exports.CONNECTED) {
                this.resetThrottleQueue.next(payload.msg || 'onSocket(logout)');
            }
            else {
                config_1.log.warn('PadchatRpc', 'onSocket() resetThrottleQueue not exist');
            }
            return;
        }
        if (payload.type === padchat_schemas_1.PadchatPayloadType.InvalidPadchatToken) {
            config_1.log.error('PadchatRpc', 'onSocket(payload.type=%s) invalid padchat token, payload=%s(%s)', padchat_schemas_1.PadchatPayloadType[payload.type], payload.type, JSON.stringify(payload));
            throw new Error(`


===========================================================================================

      The token is invalid, please use an valid token to access padchat
      If you have question about this, please check out our wiki:
      https://github.com/lijiarui/wechaty-puppet-padchat/wiki/Buy-Padchat-Token

      您使用的Token是无效的，请您联系我们获取有效Token，详情请参考这个页面：
      https://github.com/lijiarui/wechaty-puppet-padchat/wiki/%E8%B4%AD%E4%B9%B0token

============================================================================================


      `);
        }
        if (payload.type === padchat_schemas_1.PadchatPayloadType.OnlinePadchatToken) {
            config_1.log.error('PadchatRpc', 'onSocket(payload.type=%s) token is already connected with a bot, please don\'t use the same token to start multiple bot, payload=%s(%s)', padchat_schemas_1.PadchatPayloadType[payload.type], payload.type, JSON.stringify(payload));
            throw new Error(`


===========================================================================================

      The token you are using is logged in with another bot, please don't use the same
      token to start multiple bots. If you have question about this, please check
      out our wiki and contact us:
      https://github.com/lijiarui/wechaty-puppet-padchat/wiki/Buy-Padchat-Token

      您使用的Token已经登录了一个机器人，请不要使用同一个Token登录多个机器人，如果您对此有疑问，请
      参考这个页面上的联系方式联系我们：
      https://github.com/lijiarui/wechaty-puppet-padchat/wiki/%E8%B4%AD%E4%B9%B0token

============================================================================================


      `);
        }
        if (payload.type === padchat_schemas_1.PadchatPayloadType.ExpirePadchatToken) {
            config_1.log.error('PadchatRpc', 'onSocket(payload.type=%s) token is expired, please renew the token, payload=%s(%s)', padchat_schemas_1.PadchatPayloadType[payload.type], payload.type, JSON.stringify(payload));
            throw new Error(`


===========================================================================================

      The token you are using is expired, please contact us and renew this token
      Here is our wiki, you can find the contact info here:
      https://github.com/lijiarui/wechaty-puppet-padchat/wiki/Buy-Padchat-Token

      您使用的Token已经过期了，如果您想继续使用wechaty-puppet-padchat，请续费您的Token
      详情请参考以下这个页面
      https://github.com/lijiarui/wechaty-puppet-padchat/wiki/%E8%B4%AD%E4%B9%B0token

============================================================================================


      `);
        }
        if (!payload.msgId && !payload.data) {
            /**
             * Discard message that have neither msgId(Padchat API Call) nor data(Tencent Message)
             */
            if (Object.keys(payload).length === 4) {
                // {"apiName":"","data":"","msgId":"","userId":"padchat-token-zixia"}
                // just return for this message
                return;
            }
            config_1.log.silly('PadchatRpc', 'onSocket() discard payload without `msgId` and `data` for: %s', JSON.stringify(payload));
            return;
        }
        if (payload.msgId) {
            // 1. Padchat Payload
            //
            // padchatPayload:
            // {
            //     "apiName": "WXHeartBeat",
            //     "data": "%7B%22status%22%3A0%2C%22message%22%3A%22ok%22%7D",
            //     "msgId": "abc231923912983",
            //     "userId": "test"
            // }
            this.onSocketPadchat(payload);
        }
        else {
            // 2. Tencent Payload
            //
            // messagePayload:
            // {
            //   "apiName": "",
            //   "data": "XXXX",
            //   "msgId": "",
            //   "userId": "test"
            // }
            const tencentPayloadList = pure_function_helpers_1.padchatDecode(payload.data);
            if (!Array.isArray(tencentPayloadList)) {
                throw new Error('not array');
            }
            this.onSocketTencent(tencentPayloadList);
        }
        RPC_TIMEOUT_COUNTER = 0;
    }
    onSocketPadchat(padchatPayload) {
        // log.verbose('PadchatRpc', 'onSocketPadchat({apiName="%s", msgId="%s", ...})',
        //                                     padchatPayload.apiName,
        //                                     padchatPayload.msgId,
        //             )
        // log.silly('PadchatRpc', 'onSocketPadchat(%s)', JSON.stringify(padchatPayload).substr(0, 500))
        let result;
        if (padchatPayload.data) {
            result = pure_function_helpers_1.padchatDecode(padchatPayload.data);
        }
        else {
            config_1.log.silly('PadchatRpc', 'onServerMessagePadchat() discard empty payload.data for apiName: %s', padchatPayload.apiName);
            result = {};
        }
        // const jsonRpcResponse: JsonRpcPayloadResponse = {
        const jsonRpcResponse = {
            id: padchatPayload.msgId,
            jsonrpc: '2.0',
            result,
            type: 'response',
        };
        const responseText = JSON.stringify(jsonRpcResponse);
        // log.silly('PadchatRpc', 'onSocketPadchat() converted to JsonRpc payload="%s"', responseText.substr(0, 500))
        this.jsonRpc.write(responseText);
    }
    onSocketTencent(messagePayloadList) {
        // console.log('tencent messagePayloadList:', messagePayloadList)
        for (const messagePayload of messagePayloadList) {
            if (!messagePayload.msg_id) {
                // {"continue":0,"msg_type":32768,"status":1,"uin":1928023446}
                config_1.log.silly('PadchatRpc', 'onSocketTencent() discard empty message msg_id payoad: %s', JSON.stringify(messagePayload));
                continue;
            }
            // log.silly('PadchatRpc', 'onSocketTencent() messagePayload: %s', JSON.stringify(messagePayload))
            this.emit('message', messagePayload);
        }
    }
    replayTextMsg(msgId, to, text) {
        const payload = this.generateBaseMsg(msgId, to);
        payload.sub_type = padchat_schemas_1.PadchatMessageType.Text;
        payload.content = text;
        config_1.log.silly('PadchatRpc', 'replayTextMsg replaying message: %s', JSON.stringify(payload));
        this.onSocketTencent([payload]);
    }
    replayImageMsg(msgId, to, data) {
        const payload = this.generateBaseMsg(msgId, to);
        payload.sub_type = padchat_schemas_1.PadchatMessageType.Image;
        payload.data = data;
        config_1.log.silly('PadchatRpc', 'replayImageMsg replaying message: %s', JSON.stringify(payload).substr(0, 200));
        this.onSocketTencent([payload]);
    }
    replayAppMsg(msgId, to, content) {
        const payload = this.generateBaseMsg(msgId, to);
        payload.sub_type = padchat_schemas_1.PadchatMessageType.App;
        // Special conversion for Url Link message
        payload.content = `<msg>${content.trim().replace(/\n/g, '').replace(/> +</g, '><')}</msg>`;
        config_1.log.silly('PadchatRpc', 'replayAppMsg replaying message: %s', JSON.stringify(payload).substr(0, 200));
        this.onSocketTencent([payload]);
    }
    // Generate base message that sent from self
    generateBaseMsg(msgId, to) {
        const msg = {
            content: '',
            continue: padchat_schemas_1.PadchatContinue.Done,
            description: '',
            from_user: this.userId,
            msg_id: msgId,
            msg_source: '',
            msg_type: padchat_schemas_1.PadchatMessageMsgType.Five,
            status: padchat_schemas_1.PadchatMessageStatus.One,
            sub_type: padchat_schemas_1.PadchatMessageType.App,
            timestamp: new Date().getTime() / 1000,
            to_user: to,
            uin: 0
        };
        config_1.log.silly('PadchatRpc', 'generateBaseMsg(%s, %s) %s', msgId, to, JSON.stringify(msg));
        return msg;
    }
    /**
     * Init with WebSocket Server
     */
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('init');
            config_1.log.silly('PadchatRpc', 'init result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('cannot connect to WebSocket init');
            }
            return result;
        });
    }
    /**
     * Get WX block memory
     */
    WXInitialize() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PadchatRpc', 'WXInitialize()');
            const result = yield this.rpcCall('WXInitialize');
            config_1.log.silly('PadchatRpc', 'WXInitialize result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('cannot connect to WebSocket WXInitialize');
            }
            return result;
        });
    }
    WXGetQRCode() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetQRCode');
            return result;
        });
    }
    WXCheckQRCode() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXCheckQRCode');
            config_1.log.silly('PadchatRpc', 'WXCheckQRCode result: %s', JSON.stringify(result));
            if (!result) {
                throw Error('cannot connect to WebSocket WXCheckQRCode');
            }
            return result;
        });
    }
    WXHeartBeat() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXHeartBeat');
            config_1.log.silly('PadchatRpc', 'WXHeartBeat result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXHeartBeat error! canot get result from websocket server');
            }
            return result;
        });
    }
    /**
     * Load all Contact and Room
     * see issue https://github.com/lijiarui/wechaty-puppet-padchat/issues/39
     * @returns {Promise<(PadchatRoomPayload | PadchatContactPayload)[]>}
     */
    WXSyncContact() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSyncContact');
            if (!result) {
                throw Error('WXSyncContact error! canot get result from websocket server');
            }
            return result;
        });
    }
    /**
     * Generate 62 data
     *
     * 1. Call multiple times in the same session, will return the same data
     * 2. Call multiple times between sessions with the same token, will return the same data
     */
    WXGenerateWxDat() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGenerateWxDat');
            config_1.log.silly('PadchatRpc', 'WXGenerateWxDat result: %s', JSON.stringify(result));
            if (!result || !(result.data) || result.status !== 0) {
                throw Error('WXGenerateWxDat error! canot get result from websocket server');
            }
            return result.data;
        });
    }
    /**
     * Load 62 data
     * @param {string} wxData     autoData.wxData
     */
    WXLoadWxDat(wxData) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXLoadWxDat', wxData);
            config_1.log.silly('PadchatRpc', 'WXLoadWxDat result: %s', JSON.stringify(result).substr(0, 100));
            if (!result || result.status !== 0) {
                throw Error('WXLoadWxDat error! canot get result from websocket server');
            }
            return result;
        });
    }
    WXGetLoginToken() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetLoginToken');
            config_1.log.silly('PadchatRpc', 'WXGetLoginToken result: %s', JSON.stringify(result));
            if (!result || !result.token || result.status !== 0) {
                throw Error('WXGetLoginToken error! canot get result from websocket server');
            }
            return result.token;
        });
    }
    /**
     * Login with token automatically
     * @param {string} token    autoData.token
     * @returns {string} user_name | ''
     */
    WXAutoLogin(token) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXAutoLogin', token);
            config_1.log.silly('PadchatRpc', 'WXAutoLogin result: %s, type: %s', JSON.stringify(result), typeof result);
            return result;
        });
    }
    /**
     * Login with QRcode
     * @param {string} token    autoData.token
     */
    WXLoginRequest(token) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXLoginRequest', token);
            config_1.log.silly('PadchatRpc', 'WXLoginRequest result: %s, type: %s', JSON.stringify(result), typeof result);
            return result;
        });
    }
    /**
     * Send Text Message
     * @param {string} to       user_name
     * @param {string} content  text
     */
    WXSendMsg(to, content, at) {
        return __awaiter(this, void 0, void 0, function* () {
            if (to) {
                const result = yield this.rpcCall('WXSendMsg', to, content, at ? JSON.stringify(at) : '');
                if (!result || result.status !== 0) {
                    throw Error('WXSendMsg error! canot get result from websocket server');
                }
                this.replayTextMsg(result.msg_id, to, content);
                return result;
            }
            throw Error('PadchatRpc, WXSendMsg error! no to!');
        });
    }
    /**
     * Send Image Message
     * @param {string} to     user_name
     * @param {string} data   image_data
     */
    WXSendImage(to, data) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSendImage', to, data);
            this.replayImageMsg(result.msg_id, to, data);
        });
    }
    /**
     * Get contact by contact id
     * @param {any} id        user_name
     */
    WXGetContact(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetContact', id);
            if (!result) {
                throw Error('PadchatRpc, WXGetContact, cannot get result from websocket server!');
            }
            config_1.log.silly('PadchatRpc', 'WXGetContact(%s) result: %s', id, JSON.stringify(result));
            if (!result.user_name) {
                config_1.log.silly('PadchatRpc', 'WXGetContact cannot get user_name, id: %s, "%s"', id, JSON.stringify(result));
            }
            return result;
        });
    }
    /**
     * Get contact by contact id
     * @param {any} id        user_name
     */
    WXGetContactPayload(id) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!pure_function_helpers_1.isContactId(id)) { // /@chatroom$/.test(id)) {
                throw Error(`should use WXGetRoomPayload because get a room id :${id}`);
            }
            const result = yield this.WXGetContact(id);
            return result;
        });
    }
    /**
     * Get contact by contact id
     * @param {any} id        user_name
     */
    WXGetRoomPayload(id) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!pure_function_helpers_1.isRoomId(id)) { // (/@chatroom$/.test(id))) {
                throw Error(`should use WXGetContactPayload because get a contact id :${id}`);
            }
            const result = yield this.WXGetContact(id);
            if (result.member) {
                result.member = pure_function_helpers_1.padchatDecode(result.member);
            }
            return result;
        });
    }
    /**
     * Get all member of a room by room id
     * @param {any} roomId        chatroom_id
     */
    WXGetChatRoomMember(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetChatRoomMember', roomId);
            if (!result) {
                throw Error('PadchatRpc, WXGetChatRoomMember, cannot get result from websocket server!');
            }
            // roomId not exist. (or no permision?)
            // See: https://github.com/lijiarui/wechaty-puppet-padchat/issues/64#issuecomment-397319016
            if (result.status === -19) {
                return null;
            }
            /**
             * Bot quit the room, no roomId
             * See: https://github.com/lijiarui/wechaty-puppet-padchat/issues/38
             * {"chatroom_id":0,"count":0,"member":"null\n","message":"","status":0,"user_name":""}
             */
            if (Object.keys(result).length === 0 || (result.count === 0 && result.status === 0 && result.chatroom_id === 0)) {
                return null;
            }
            config_1.log.silly('PadchatRpc', 'WXGetChatRoomMember() result: %s', JSON.stringify(result).substr(0, 500));
            // 00:40:44 SILL PadchatRpc WXGetChatRoomMember() result: {"chatroom_id":0,"count":0,"member":"null\n","message":"","status":0,"user_name":""}
            try {
                const tryMemberList = pure_function_helpers_1.padchatDecode(result.member);
                if (Array.isArray(tryMemberList)) {
                    result.member = tryMemberList;
                }
                else if (tryMemberList !== null) {
                    config_1.log.warn('PadchatRpc', 'WXGetChatRoomMember(%s) member is neither array nor null: %s', roomId, JSON.stringify(result.member));
                    // throw Error('faild to parse chatroom member!')
                    result.member = [];
                }
            }
            catch (e) {
                config_1.log.warn('PadchatRpc', 'WXGetChatRoomMember(%s) result.member decode error: %s', roomId, e);
                result.member = [];
            }
            return result;
        });
    }
    /**
     * Login successfully by qrcode
     * @param {any} username        user_name
     * @param {any} password  password
     */
    WXQRCodeLogin(username, password) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXQRCodeLogin', username, password);
            if (!result) {
                throw Error(`PadchatRpc, WXQRCodeLogin, unknown error, data: ${JSON.stringify(result)}`);
            }
            if (result.status === 0) {
                config_1.log.silly('PadchatRpc', 'WXQRCodeLogin, login successfully!');
                return result;
            }
            if (result.status === -3) {
                throw Error('PadchatRpc, WXQRCodeLogin, wrong user_name or password');
            }
            else if (result.status === -301) {
                config_1.log.warn('PadchatRpc', 'WXQRCodeLogin, redirect 301');
                return this.WXQRCodeLogin(username, password);
            }
            throw Error(`PadchatRpc, WXQRCodeLogin,
    If the status is -106, please refer to https://github.com/botorange/puppet-padchat-patch for a temporary solution
    unknown status: ${result.status}, for username: ${username}`);
        });
    }
    WXSetUserRemark(id, remark) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSetUserRemark', id, remark);
            config_1.log.silly('PadchatRpc', 'WXSetUserRemark result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSetUserRemark error! canot get result from websocket server');
            }
            return result;
        });
    }
    WXDeleteChatRoomMember(roomId, contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXDeleteChatRoomMember', roomId, contactId);
            config_1.log.silly('PadchatRpc', 'WXDeleteChatRoomMember result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXDeleteChatRoomMember error! canot get result from websocket server');
            }
            return result;
        });
    }
    WXAddChatRoomMember(roomId, contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXAddChatRoomMember', roomId, contactId);
            config_1.log.silly('PadchatRpc', 'WXAddChatRoomMember result: %s', JSON.stringify(result));
            if (!result) {
                throw Error('WXAddChatRoomMember error! canot get result from websocket server');
            }
            if (result.status === padchat_rpc_type_1.WXRoomAddTypeStatus.Done) {
                // see more in WXAddChatRoomMemberType
                if (/OK/i.test(result.message)) {
                    return 0;
                }
                config_1.log.warn('PadchatRpc', 'WXAddChatRoomMember() status = 0 but message is not OK: ' + result.message);
                return -1;
            }
            /**
             * see https://github.com/lijiarui/wechaty-puppet-padchat/issues/70
             * If room member more than 40
             * Need call `WXInviteChatRoomMember` instead `WXAddChatRoomMember`
             */
            if (result.status === padchat_rpc_type_1.WXRoomAddTypeStatus.NeedInvite) {
                config_1.log.silly('PadchatRpc', 'WXAddChatRoomMember change to WXInviteChatRoomMember');
                return this.WXInviteChatRoomMember(roomId, contactId);
            }
            if (result.status === padchat_rpc_type_1.WXRoomAddTypeStatus.InviteConfirm) {
                // result: {"message":"","status":-2028}
                // May be the owner has see not allow other people to join in the room (群聊邀请确认)
                config_1.log.warn('PadchatRpc', 'WXAddChatRoomMember failed! maybe owner open the should confirm first to invited others to join in the room.');
            }
            return result.status;
        });
    }
    /**
     * When member more than 40, use WXInviteChatRoomMember
     * When member less than 40, use WXAddChatRoomMember
     *
     * @param {string} roomId
     * @param {string} contactId
     * @returns {Promise<number>}
     */
    WXInviteChatRoomMember(roomId, contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            // TODO:
            // Change `WXAddChatRoomMemberType` to `WXInviteChatRoomMemberType`
            const result = yield this.rpcCall('WXInviteChatRoomMember', roomId, contactId);
            config_1.log.silly('PadchatRpc', 'WXInviteChatRoomMember result: %s', JSON.stringify(result));
            if (!result) {
                throw Error('WXInviteChatRoomMember error! canot get result from websocket server');
            }
            if (result.status === padchat_rpc_type_1.WXRoomAddTypeStatus.Done) {
                // see more in WXAddChatRoomMemberType
                if (/OK/i.test(result.message)) {
                    return 0;
                }
                config_1.log.warn('PadchatRpc', 'WXInviteChatRoomMember() status = 0 but message is not OK: ' + result.message);
                return -1;
            }
            // TODO
            // Should check later
            if (result.status === -padchat_rpc_type_1.WXRoomAddTypeStatus.InviteConfirm) {
                // result: {"message":"","status":-2028}
                // May be the owner has see not allow other people to join in the room (群聊邀请确认)
                config_1.log.warn('PadchatRpc', 'WXInviteChatRoomMember failed! maybe owner open the should confirm first to invited others to join in the room.');
            }
            return result.status;
        });
    }
    WXSetChatroomName(roomId, topic) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSetChatroomName', roomId, topic);
            config_1.log.silly('PadchatRpc', 'WXSetChatroomName result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSetChatroomName error! canot get result from websocket server');
            }
            return result;
        });
    }
    WXQuitChatRoom(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXQuitChatRoom', roomId);
            config_1.log.silly('PadchatRpc', 'WXQuitChatRoom result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXQuitChatRoom error! canot get result from websocket server');
            }
            return result;
        });
    }
    WXAddUser(strangerV1, strangerV2, type, verify) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXAddUser', strangerV1, strangerV2, type, verify);
            config_1.log.silly('PadchatRpc', 'WXAddUser result: %s', JSON.stringify(result));
            return result;
        });
    }
    WXAcceptUser(stranger, ticket) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXAcceptUser', stranger, ticket);
            config_1.log.silly('PadchatRpc', 'WXAcceptUser result: %s', JSON.stringify(result));
            return result;
        });
    }
    WXLogout() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXLogout');
            config_1.log.silly('PadchatRpc', 'WXLogout result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXLogout error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO: check any
    WXSendMoments(text) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSendMoments', text);
            config_1.log.silly('PadchatRpc', 'WXSendMoments result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSendMoments error! canot get result from websocket server');
            }
            return result;
        });
    }
    // Sync Chat Message
    WXSyncMessage() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                this.rpcCall('WXSyncMessage').then((result) => {
                    config_1.log.silly('PadchatRpc', 'WXSyncMessage result: %s', JSON.stringify(result));
                    if (!result || !result[0] || (result[0].status !== 1 && result[0].status !== -1)) {
                        resolve();
                    }
                    else {
                        const tencentPayloadList = pure_function_helpers_1.padchatDecode(result);
                        if (!Array.isArray(tencentPayloadList)) {
                            config_1.log.warn('PadchatRpc', 'WXSyncMessage() parsed data is not an array: %s', JSON.stringify(tencentPayloadList));
                            reject('parsed-data-not-array');
                        }
                        this.onSocketTencent(tencentPayloadList);
                        resolve();
                    }
                }).catch(reject);
                const timer = setTimeout(() => {
                    clearTimeout(timer);
                    reject('timeout');
                }, config_1.CON_TIME_OUT);
            });
        });
    }
    // This function is used for add new friends by id
    WXSearchContact(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSearchContact', id);
            config_1.log.silly('PadchatRpc', 'WXSearchContact result: %s', JSON.stringify(result));
            if ((!result) && (result.status !== padchat_rpc_type_1.WXSearchContactTypeStatus.Searchable) && (result.status !== padchat_rpc_type_1.WXSearchContactTypeStatus.UnSearchable)) {
                throw Error('WXSearchContact error! canot get result from websocket server');
            }
            if (result.status === padchat_rpc_type_1.WXSearchContactTypeStatus.Searchable) {
                config_1.log.verbose('PadchatRpc', 'WXSearchContact wxid: %s can be searched', id);
            }
            if (result.status === padchat_rpc_type_1.WXSearchContactTypeStatus.UnSearchable) {
                config_1.log.verbose('PadchatRpc', 'WXSearchContact wxid: %s cannot be searched', id);
                throw Error(`ContactId ${id} is not searchable`);
            }
            return result;
        });
    }
    // TODO: check any
    // send hello to strange
    WXSayHello(stranger, text) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSayHello', stranger, text);
            config_1.log.silly('PadchatRpc', 'WXSayHello , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSayHello , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO: check any
    // delete friend
    WXDeleteUser(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXDeleteUser', id);
            config_1.log.silly('PadchatRpc', 'WXDeleteUser , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXDeleteUser , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    WXCreateChatRoom(userList) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXCreateChatRoom', JSON.stringify(userList));
            config_1.log.silly('PadchatRpc', 'WXCreateChatRoom(userList.length=%d) = "%s"', userList.length, JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXCreateChatRoom , stranger,error! canot get result from websocket server');
            }
            return pure_function_helpers_1.stripBugChatroomId(result.user_name);
        });
    }
    // TODO: check any
    // TODO: receive red-packet automatically
    // red_packet: see this from the recived message
    WXReceiveRedPacket(redPacket) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXReceiveRedPacket', redPacket);
            config_1.log.silly('PadchatRpc', 'WXReceiveRedPacket , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXReceiveRedPacket , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO: check any
    // TODO: 查看红包
    WXQueryRedPacket(redPacket) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXQueryRedPacket', redPacket);
            config_1.log.silly('PadchatRpc', 'WXQueryRedPacket , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXQueryRedPacket , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO: check any
    // TODO: 打开红包
    WXOpenRedPacket(redPacket) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXOpenRedPacket', redPacket);
            config_1.log.silly('PadchatRpc', 'WXOpenRedPacket , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXOpenRedPacket , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO: check any
    // set user avatar, param image is the same as send [WXSendImage]
    WXSetHeadImage(image) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXOpenRedPacket', image);
            config_1.log.silly('PadchatRpc', 'WXOpenRedPacket , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXOpenRedPacket , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO: check any
    // nick_name	昵称
    // signature	签名
    // sex			性别，1男，2女
    // country		国家，CN
    // provincia	地区，省，Zhejiang
    // city			城市，Hangzhou
    WXSetUserInfo(nickeName, signature, sex, country, provincia, city) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSetUserInfo', nickeName, signature, sex, country, provincia, city);
            config_1.log.silly('PadchatRpc', 'WXSetUserInfo , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSetUserInfo , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO: check any
    // 查看附近的人
    // longitude	浮点数，经度
    // latitude		浮点数，维度
    WXGetPeopleNearby(longitude, latitude) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetPeopleNearby', longitude, latitude);
            config_1.log.silly('PadchatRpc', 'WXGetPeopleNearby , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXGetPeopleNearby , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO: check any
    // 获取公众号信息 id: gh_xxxx
    WXGetSubscriptionInfo(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetSubscriptionInfo', id);
            config_1.log.silly('PadchatRpc', 'WXGetSubscriptionInfo , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXGetSubscriptionInfo , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO: check any
    // 执行公众号菜单操作
    // id			      公众号用户名
    // OrderId			菜单id，通过WXGetSubscriptionInfo获取, 原来叫id
    // OrderKey			菜单key，通过WXGetSubscriptionInfo获取, 原来叫key
    WXSubscriptionCommand(id, orderId, orderKey) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSubscriptionCommand', id, orderId, orderKey);
            config_1.log.silly('PadchatRpc', 'WXSubscriptionCommand , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSubscriptionCommand , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 公众号中获取url访问token, 给下面的函数使用[WXRequestUrl]
    // user			公众号用户名
    // url			访问的链接
    WXGetRequestToken(id, url) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetRequestToken', id, url);
            config_1.log.silly('PadchatRpc', 'WXGetRequestToken , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXGetRequestToken , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 公众号中访问url
    // url			url地址
    // key			token中的key
    // uin		  token中的uin
    WXRequestUrl(url, key, uin) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXRequestUrl', url, key, uin);
            config_1.log.silly('PadchatRpc', 'WXRequestUrl , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXRequestUrl , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 设置微信ID
    WXSetWeChatID(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSetWeChatID', id);
            config_1.log.silly('PadchatRpc', 'WXSetWeChatID , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSetWeChatID , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 查看转账信息
    // transfer		转账消息
    WXTransferQuery(transfer) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXTransferQuery', transfer);
            config_1.log.silly('PadchatRpc', 'WXTransferQuery , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXTransferQuery , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 接受转账
    // transfer		转账消息
    WXTransferOperation(transfer) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXTransferOperation', transfer);
            config_1.log.silly('PadchatRpc', 'WXTransferOperation , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXTransferOperation , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 获取消息图片 （应该是查看原图）
    // msg			收到的整个图片消息
    WXGetMsgImage(msg) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetMsgImage', msg);
            config_1.log.silly('PadchatRpc', 'WXGetMsgImage , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXGetMsgImage , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 获取消息视频 （应该是查看视频）
    // msg			收到的整个视频消息
    WXGetMsgVideo(msg) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetMsgVideo', msg);
            config_1.log.silly('PadchatRpc', 'WXGetMsgVideo , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXGetMsgVideo , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 获取消息语音(语音消息大于20秒时通过该接口获取)
    // msg			收到的整个视频消息
    WXGetMsgVoice(msg) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetMsgVoice', msg);
            config_1.log.silly('PadchatRpc', 'WXGetMsgVoice ,result: %s', JSON.stringify(result).substr(0, 100));
            if (!result || result.status !== 0) {
                throw Error('WXGetMsgVoice , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 搜索公众号
    // id			公众号用户名: gh_xxx
    WXWebSearch(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXWebSearch', id);
            config_1.log.silly('PadchatRpc', 'WXWebSearch , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXWebSearch , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 分享名片
    // user			对方用户名
    // id			    名片的微信id
    // caption		名片的标题
    WXShareCard(user, id, caption) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXShareCard', user, id, caption);
            config_1.log.silly('PadchatRpc', 'WXShareCard , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXShareCard , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 重置同步信息
    WXSyncReset() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSyncReset');
            config_1.log.silly('PadchatRpc', 'WXSyncReset , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSyncReset , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 扫描二维码获取信息
    // path			本地二维码图片全路径
    WXQRCodeDecode(path) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXQRCodeDecode', path);
            config_1.log.silly('PadchatRpc', 'WXQRCodeDecode , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXQRCodeDecode , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 朋友圈上传图片获取url
    // image			图片数据, 和发送消息的image 是一样的base64 串
    WXSnsUpload(image) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSnsUpload', image);
            config_1.log.silly('PadchatRpc', 'WXSnsUpload , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSnsUpload , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 获取朋友圈消息详情(例如评论)
    // id			朋友圈消息id
    WXSnsObjectDetail(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSnsObjectDetail', id);
            config_1.log.silly('PadchatRpc', 'WXSnsObjectDetail , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSnsObjectDetail , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 朋友圈操作(删除朋友圈，删除评论，取消赞)
    // id			       朋友圈消息id
    // type			     操作类型,1为删除朋友圈，4为删除评论，5为取消赞
    // comment		   当type为4时，对应删除评论的id，通过WXSnsObjectDetail接口获取。当type为其他值时，comment不可用，置为0。
    // commentType	 评论类型,当删除评论时可用，2或者3.(规律未知)
    WXSnsObjectOp(id, type, comment, commentType) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSnsObjectOp', id, type, comment, commentType);
            config_1.log.silly('PadchatRpc', 'WXSnsObjectOp , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSnsObjectOp , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 朋友圈消息评论
    // user			  对方用户名
    // id			    朋友圈消息id
    // content		评论内容
    // replyId		回复的id    //如果想回复某人的评论，就加上他的comment_id  否则就用0
    WXSnsComment(user, id, content, replyId) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSnsComment', user, id, content, replyId);
            config_1.log.silly('PadchatRpc', 'WXSnsComment , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSnsComment , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 获取好友朋友圈信息
    // user			对方用户名
    // id			    获取到的最后一次的id，第一次调用设置为空
    WXSnsUserPage(user, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSnsUserPage', user, id);
            config_1.log.silly('PadchatRpc', 'WXSnsUserPage , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSnsUserPage , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 获取朋友圈动态
    // id			    获取到的最后一次的id，第一次调用设置为空
    WXSnsTimeline(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSnsTimeline', id);
            config_1.log.silly('PadchatRpc', 'WXSnsTimeline , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSnsTimeline , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 发送APP消息(分享应用或者朋友圈链接等)
    // user			对方用户名
    // content		消息内容(整个消息结构<appmsg xxxxxxxxx>) 参考收到的MsgType
    WXSendAppMsg(user, content) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSendAppMsg', user, content);
            config_1.log.silly('PadchatRpc', 'WXSendAppMsg , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSendAppMsg , stranger,error! canot get result from websocket server');
            }
            this.replayAppMsg(result.msg_id, user, content);
            return result;
        });
    }
    // TODO check any
    // 同步收藏消息(用户获取收藏对象的id)
    // key			同步的key，第一次调用设置为空。
    WXFavSync(key) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXFavSync', key);
            config_1.log.silly('PadchatRpc', 'WXFavSync , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXFavSync , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 添加收藏
    // fav_object	收藏对象结构(<favitem type=5xxxxxx)
    WXFavAddItem(favObject) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXFavAddItem', favObject);
            config_1.log.silly('PadchatRpc', 'WXFavAddItem , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXFavAddItem , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 获取收藏对象的详细信息
    // id			收藏对象id
    WXFavGetItem(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXFavGetItem', id);
            config_1.log.silly('PadchatRpc', 'WXFavGetItem , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXFavGetItem , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 删除收藏对象
    // id			收藏对象id
    WXFavDeleteItem(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXFavDeleteItem', id);
            config_1.log.silly('PadchatRpc', 'WXFavDeleteItem , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXFavDeleteItem , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 获取所有标签列表
    WXGetContactLabelList() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetContactLabelList');
            config_1.log.silly('PadchatRpc', 'WXGetContactLabelList , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXGetContactLabelList , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 添加标签到列表
    // label			标签内容
    WXAddContactLabel(label) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXAddContactLabel', label);
            config_1.log.silly('PadchatRpc', 'WXAddContactLabel , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXAddContactLabel , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 从列表删除标签
    // id			标签id
    WXDeleteContactLabel(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXDeleteContactLabel', id);
            config_1.log.silly('PadchatRpc', 'WXDeleteContactLabel , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXDeleteContactLabel , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 从列表删除标签
    // user			用户名
    // id			    标签id
    WXSetContactLabel(user, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSetContactLabel', user, id);
            config_1.log.silly('PadchatRpc', 'WXSetContactLabel , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSetContactLabel , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 获取用户二维码(自己或者已加入的群)
    // user			用户名
    // style			是否使用风格化二维码
    WXGetUserQRCode(user, style) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXGetUserQRCode', user, style);
            config_1.log.silly('PadchatRpc', 'WXGetUserQRCode , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0 || !result.qr_code) {
                throw Error('WXGetUserQRCode, error: canot get result from websocket server');
            }
            // console.log('WXGetUserQRCode result:', result)
            return result.qr_code;
        });
    }
    // TODO check any
    // AppMsg上传数据
    // data			和发送图片的data 类似
    WXUploadAppAttach(data) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXUploadAppAttach', data);
            config_1.log.silly('PadchatRpc', 'WXUploadAppAttach , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXUploadAppAttach , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 发送语音消息(微信silk格式语音)
    // user			对方用户名
    // voice_data	语音数据
    // voice_size	语音大小 (应该不需要)
    // voice_time	语音时间(毫秒，最大60 * 1000)
    WXSendVoice(user, data, time) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSendVoice', user, data, time);
            config_1.log.silly('PadchatRpc', 'WXSendVoice , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSendVoice , stranger,error! canot get result from websocket server');
            }
            // TODO: replay voice message
            return result;
        });
    }
    // TODO check any
    // 同步朋友圈动态(好友评论或点赞自己朋友圈的消息)
    // key		同步key
    WXSnsSync(key) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSnsSync', key);
            config_1.log.silly('PadchatRpc', 'WXSnsSync , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSnsSync , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 群发消息
    // user			用户名json数组 ["AB1","AC2","AD3"]
    // content		消息内容
    WXMassMessage(userList, content) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXMassMessage', JSON.stringify(userList), content);
            config_1.log.silly('PadchatRpc', 'WXMassMessage , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXMassMessage , stranger,error! canot get result from websocket server');
            }
            return result;
        });
    }
    // TODO check any
    // 设置群公告
    // chatroom	群id
    // content	    内容
    WXSetChatroomAnnouncement(chatroom, content) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.rpcCall('WXSetChatroomAnnouncement', chatroom, content);
            config_1.log.silly('PadchatRpc', 'WXSetChatroomAnnouncement , stranger,result: %s', JSON.stringify(result));
            if (!result || result.status !== 0) {
                throw Error('WXSetChatroomAnnouncement , stranger,error! canot get result from websocket server');
            }
            console.log('WXSetChatroomAnnouncement result:', result);
            return result;
        });
    }
}
exports.PadchatRpc = PadchatRpc;
//# sourceMappingURL=padchat-rpc.js.map