"use strict";
/**
 *   Wechaty - https://github.com/chatie/wechaty
 *
 *   @copyright 2016-2018 Huan LI <zixia@zixia.net>
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const array_flatten_1 = __importDefault(require("array-flatten"));
const lru_cache_1 = __importDefault(require("lru-cache"));
const file_box_1 = require("file-box");
const wechaty_puppet_1 = require("wechaty-puppet");
const pure_function_helpers_1 = require("./pure-function-helpers");
const config_1 = require("./config");
const padchat_manager_1 = require("./padchat-manager");
const padchat_schemas_1 = require("./padchat-schemas");
const padchat_rpc_type_1 = require("./padchat-rpc.type");
const app_message_generator_1 = require("./pure-function-helpers/app-message-generator");
const message_emoji_payload_parser_1 = require("./pure-function-helpers/message-emoji-payload-parser");
let PADCHAT_COUNTER = 0; // PuppetPadchat Instance Counter
class PuppetPadchat extends wechaty_puppet_1.Puppet {
    constructor(options = {}) {
        super(Object.assign({ timeout: 60 * 4 }, options));
        this.options = options;
        const lruOptions = {
            max: 1000,
            // length: function (n) { return n * 2},
            dispose(key, val) {
                config_1.log.silly('PuppetPadchat', 'constructor() lruOptions.dispose(%s, %s)', key, JSON.stringify(val));
            },
            maxAge: 1000 * 60 * 60,
        };
        this.cachePadchatMessagePayload = new lru_cache_1.default(lruOptions);
        this.padchatCounter = PADCHAT_COUNTER++;
        if (this.padchatCounter > 0) {
            if (!this.options.token) {
                throw new Error([
                    'You need to specify `token` when constructor PuppetPadchat becasue you have more than one instance. ',
                    'see: https://github.com/Chatie/wechaty/issues/1367',
                ].join(''));
            }
        }
    }
    toString() {
        const text = super.toString();
        return text + `/PuppetPadchat#${this.padchatCounter}`;
    }
    ding(data) {
        config_1.log.verbose('PuppetPadchat', 'ding(%s)', data || '');
        // TODO: do some internal health check inside this.padchatManager
        if (!this.padchatManager) {
            this.emit('error', new Error('no padchat Manager'));
            return;
        }
        this.padchatManager.ding(data);
        return;
    }
    startWatchdog() {
        config_1.log.verbose('PuppetPadchat', 'startWatchdog()');
        if (!this.padchatManager) {
            throw new Error('no padchat manager');
        }
        // clean the dog because this could be re-inited
        // this.watchdog.removeAllListeners()
        /**
         * Use manager's heartbeat to feed dog
         */
        this.padchatManager.on('heartbeat', (data) => {
            config_1.log.silly('PuppetPadchat', 'startWatchdog() padchatManager.on(heartbeat)');
            this.emit('watchdog', {
                data,
            });
        });
        this.emit('watchdog', {
            data: 'inited',
            type: 'startWatchdog()',
        });
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', `start() with ${this.memory.name}`);
            if (this.state.on()) {
                config_1.log.warn('PuppetPadchat', 'start() already on(pending)?');
                yield this.state.ready('on');
                return;
            }
            /**
             * state has two main state: ON / OFF
             * ON (pending)
             * OFF (pending)
             */
            this.state.on('pending');
            const manager = this.padchatManager = new padchat_manager_1.PadchatManager({
                endpoint: this.options.endpoint || config_1.WECHATY_PUPPET_PADCHAT_ENDPOINT,
                memory: this.memory,
                token: this.options.token || config_1.padchatToken(),
            });
            yield this.startManager(manager);
            yield this.startWatchdog();
            this.state.on(true);
        });
    }
    login(selfId) {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            yield _super("login").call(this, selfId);
            yield this.padchatManager.syncContactsAndRooms();
        });
    }
    startManager(manager) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'startManager()');
            if (this.state.off()) {
                throw new Error('startManager() state is off');
            }
            manager.removeAllListeners();
            manager.on('error', e => this.emit('error', e));
            manager.on('scan', (qrcode, status, data) => this.emit('scan', qrcode, status, data));
            manager.on('login', (userId) => this.login(userId));
            manager.on('message', (rawPayload) => this.onPadchatMessage(rawPayload));
            manager.on('logout', () => this.logout(true));
            manager.on('dong', (data) => this.emit('dong', data));
            manager.on('ready', () => this.emit('ready'));
            manager.on('reset', (reason) => __awaiter(this, void 0, void 0, function* () {
                config_1.log.warn('PuppetPadchat', 'startManager() manager.on(reset) for %s. Restarting PuppetPadchat ... ', reason);
                // Puppet Base class will deal with this RESET event for you.
                yield this.emit('reset', reason);
            }));
            manager.on('reconnect', (msg) => __awaiter(this, void 0, void 0, function* () {
                config_1.log.verbose('PuppetPadchat', 'startManager() manager.on(reconnect) for %s', msg);
                // Slightly delay the reconnect after disconnected from the server
                yield new Promise(r => setTimeout(r, 2000));
                yield manager.reconnect();
            }));
            yield manager.start();
        });
    }
    onPadchatMessage(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'onPadchatMessage({id=%s, type=%s(%s)})', rawPayload.msg_id, padchat_schemas_1.PadchatMessageType[rawPayload.sub_type], rawPayload.msg_type);
            /**
             * 0. Discard messages when not logged in
             */
            if (!this.id) {
                config_1.log.warn('PuppetPadchat', 'onPadchatMessage(%s) discarded message because puppet is not logged-in', JSON.stringify(rawPayload));
                return;
            }
            /**
             * 1. Sometimes will get duplicated same messages from rpc, drop the same message from here.
             */
            if (this.cachePadchatMessagePayload.has(rawPayload.msg_id)) {
                config_1.log.silly('PuppetPadchat', 'onPadchatMessage(id=%s) duplicate message: %s', rawPayload.msg_id, JSON.stringify(rawPayload).substr(0, 500));
                return;
            }
            /**
             * 2. Save message for future usage
             */
            this.cachePadchatMessagePayload.set(rawPayload.msg_id, rawPayload);
            /**
             * 3. Check for Different Message Types
             */
            switch (rawPayload.sub_type) {
                case padchat_schemas_1.PadchatMessageType.VerifyMsg:
                    this.emit('friendship', rawPayload.msg_id);
                    break;
                case padchat_schemas_1.PadchatMessageType.Recalled:
                    /**
                     * When someone joined the room invited by Bot,
                     * the bot will receive a `recall-able` message for room event
                     *
                     * { content: '12740017638@chatroom:\n<sysmsg type="delchatroommember">\n\t<delchatroommember>\n\t\t<plain>
                     *            <![CDATA[You invited 卓桓、Zhuohuan, 太阁_传话助手, 桔小秘 to the group chat.   ]]></plain>...,
                     *  sub_type: 10002}
                     */
                    yield Promise.all([
                        this.onPadchatMessageRoomEventJoin(rawPayload),
                    ]);
                    break;
                case padchat_schemas_1.PadchatMessageType.Sys:
                    yield Promise.all([
                        this.onPadchatMessageFriendshipEvent(rawPayload),
                        ////////////////////////////////////////////////
                        this.onPadchatMessageRoomEventJoin(rawPayload),
                        this.onPadchatMessageRoomEventLeave(rawPayload),
                        this.onPadchatMessageRoomEventTopic(rawPayload),
                    ]);
                    break;
                case padchat_schemas_1.PadchatMessageType.App:
                    yield Promise.all([
                        this.onPadchatMessageRoomInvitation(rawPayload),
                    ]);
                    break;
                case padchat_schemas_1.PadchatMessageType.Emoticon:
                case padchat_schemas_1.PadchatMessageType.Image:
                case padchat_schemas_1.PadchatMessageType.MicroVideo:
                case padchat_schemas_1.PadchatMessageType.Video:
                // TODO: the above types are field type
                default:
                    this.emit('message', rawPayload.msg_id);
                    break;
            }
        });
    }
    onPadchatMessageRoomInvitation(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'onPadchatMessageRoomInvitation(%s)', rawPayload);
            const roomInviteEvent = yield pure_function_helpers_1.roomInviteEventMessageParser(rawPayload);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            if (roomInviteEvent) {
                yield this.padchatManager.saveRoomInvitationRawPayload(roomInviteEvent);
                this.emit('room-invite', roomInviteEvent.msgId);
            }
            else {
                this.emit('message', rawPayload.msg_id);
            }
        });
    }
    /**
     * Look for room join event
     */
    onPadchatMessageRoomEventJoin(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'onPadchatMessageRoomEventJoin({id=%s})', rawPayload.msg_id);
            const roomJoinEvent = yield pure_function_helpers_1.roomJoinEventMessageParser(rawPayload);
            if (roomJoinEvent) {
                const inviteeNameList = roomJoinEvent.inviteeNameList;
                const inviterName = roomJoinEvent.inviterName;
                const roomId = roomJoinEvent.roomId;
                config_1.log.silly('PuppetPadchat', 'onPadchatMessageRoomEventJoin() roomJoinEvent="%s"', JSON.stringify(roomJoinEvent));
                const inviteeIdList = yield config_1.retry((retryException, attempt) => __awaiter(this, void 0, void 0, function* () {
                    config_1.log.verbose('PuppetPadchat', 'onPadchatMessageRoomEvent({id=%s}) roomJoin retry(attempt=%d)', attempt);
                    const tryIdList = array_flatten_1.default(yield Promise.all(inviteeNameList.map(inviteeName => this.roomMemberSearch(roomId, inviteeName))));
                    if (tryIdList.length) {
                        return tryIdList;
                    }
                    if (!this.padchatManager) {
                        throw new Error('no manager');
                    }
                    /**
                     * Set Cache Dirty
                     */
                    yield this.roomMemberPayloadDirty(roomId);
                    return retryException(new Error('roomMemberSearch() not found'));
                })).catch(e => {
                    config_1.log.warn('PuppetPadchat', 'onPadchatMessageRoomEvent({id=%s}) roomJoin retry() fail: %s', e.message);
                    return [];
                });
                const inviterIdList = yield this.roomMemberSearch(roomId, inviterName);
                if (inviterIdList.length < 1) {
                    throw new Error('no inviterId found');
                }
                else if (inviterIdList.length > 1) {
                    config_1.log.warn('PuppetPadchat', 'onPadchatMessageRoomEvent() case PadchatMesssageSys: inviterId found more than 1, use the first one.');
                }
                const inviterId = inviterIdList[0];
                /**
                 * Set Cache Dirty
                 */
                yield this.roomMemberPayloadDirty(roomId);
                yield this.roomPayloadDirty(roomId);
                this.emit('room-join', roomId, inviteeIdList, inviterId);
            }
        });
    }
    /**
     * Look for room leave event
     */
    onPadchatMessageRoomEventLeave(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'onPadchatMessageRoomEventLeave({id=%s})', rawPayload.msg_id);
            const roomLeaveEvent = pure_function_helpers_1.roomLeaveEventMessageParser(rawPayload);
            if (roomLeaveEvent) {
                const leaverNameList = roomLeaveEvent.leaverNameList;
                const removerName = roomLeaveEvent.removerName;
                const roomId = roomLeaveEvent.roomId;
                config_1.log.silly('PuppetPadchat', 'onPadchatMessageRoomEventLeave() roomLeaveEvent="%s"', JSON.stringify(roomLeaveEvent));
                const leaverIdList = array_flatten_1.default(yield Promise.all(leaverNameList.map(leaverName => this.roomMemberSearch(roomId, leaverName))));
                const removerIdList = yield this.roomMemberSearch(roomId, removerName);
                if (removerIdList.length < 1) {
                    throw new Error('no removerId found');
                }
                else if (removerIdList.length > 1) {
                    config_1.log.warn('PuppetPadchat', 'onPadchatMessage() case PadchatMesssageSys: removerId found more than 1, use the first one.');
                }
                const removerId = removerIdList[0];
                if (!this.padchatManager) {
                    throw new Error('no padchatManager');
                }
                /**
                 * Set Cache Dirty
                 */
                yield this.roomMemberPayloadDirty(roomId);
                yield this.roomPayloadDirty(roomId);
                this.emit('room-leave', roomId, leaverIdList, removerId);
            }
        });
    }
    /**
     * Look for room topic event
     */
    onPadchatMessageRoomEventTopic(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'onPadchatMessageRoomEventTopic({id=%s})', rawPayload.msg_id);
            const roomTopicEvent = pure_function_helpers_1.roomTopicEventMessageParser(rawPayload);
            if (roomTopicEvent) {
                const changerName = roomTopicEvent.changerName;
                const newTopic = roomTopicEvent.topic;
                const roomId = roomTopicEvent.roomId;
                config_1.log.silly('PuppetPadchat', 'onPadchatMessageRoomEventTopic() roomTopicEvent="%s"', JSON.stringify(roomTopicEvent));
                const roomOldPayload = yield this.roomPayload(roomId);
                const oldTopic = roomOldPayload.topic;
                const changerIdList = yield this.roomMemberSearch(roomId, changerName);
                if (changerIdList.length < 1) {
                    throw new Error('no changerId found');
                }
                else if (changerIdList.length > 1) {
                    config_1.log.warn('PuppetPadchat', 'onPadchatMessage() case PadchatMesssageSys: changerId found more than 1, use the first one.');
                }
                const changerId = changerIdList[0];
                if (!this.padchatManager) {
                    throw new Error('no padchatManager');
                }
                /**
                 * Set Cache Dirty
                 */
                yield this.roomPayloadDirty(roomId);
                this.emit('room-topic', roomId, newTopic, oldTopic, changerId);
            }
        });
    }
    onPadchatMessageFriendshipEvent(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'onPadchatMessageFriendshipEvent({id=%s})', rawPayload.msg_id);
            /**
             * 1. Look for friendship confirm event
             */
            const friendshipConfirmContactId = pure_function_helpers_1.friendshipConfirmEventMessageParser(rawPayload);
            /**
             * 2. Look for friendship receive event
             */
            const friendshipReceiveContactId = yield pure_function_helpers_1.friendshipReceiveEventMessageParser(rawPayload);
            /**
             * 3. Look for friendship verify event
             */
            const friendshipVerifyContactId = pure_function_helpers_1.friendshipVerifyEventMessageParser(rawPayload);
            if (friendshipConfirmContactId
                || friendshipReceiveContactId
                || friendshipVerifyContactId) {
                // Maybe load contact here since we know a new friend is added
                this.emit('friendship', rawPayload.msg_id);
            }
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'stop()');
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            if (this.state.off()) {
                config_1.log.warn('PuppetPadchat', 'stop() is called on a OFF puppet. await ready(off) and return.');
                yield this.state.ready('off');
                return;
            }
            this.state.off('pending');
            // this.watchdog.sleep()
            yield this.logout(true);
            yield this.padchatManager.stop();
            this.padchatManager.removeAllListeners();
            this.padchatManager = undefined;
            this.state.off(true);
        });
    }
    logout(shallow = false) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'logout()');
            if (!this.id) {
                config_1.log.warn('PuppetPadchat', 'logout() this.id not exist');
                return;
            }
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            this.emit('logout', this.id); // becore we will throw above by logonoff() when this.user===undefined
            this.id = undefined;
            if (!shallow) {
                yield this.padchatManager.WXLogout();
            }
            yield this.padchatManager.logout();
        });
    }
    contactAlias(contactId, alias) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'contactAlias(%s, %s)', contactId, alias);
            if (typeof alias === 'undefined') {
                const payload = yield this.contactPayload(contactId);
                return payload.alias || '';
            }
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            yield this.padchatManager.WXSetUserRemark(contactId, alias || '');
            return;
        });
    }
    contactValidate(contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'contactValid(%s)', contactId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            try {
                yield this.padchatManager.contactRawPayload(contactId);
                return true;
            }
            catch (e) {
                return false;
            }
        });
    }
    contactList() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'contactList()');
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const contactIdList = this.padchatManager.getContactIdList();
            return contactIdList;
        });
    }
    contactAvatar(contactId, file) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'contactAvatar(%s%s)', contactId, file ? (', ' + file.name) : '');
            /**
             * 1. set avatar for user self
             */
            if (file) {
                if (contactId !== this.selfId()) {
                    throw new Error('can not set avatar for others');
                }
                if (!this.padchatManager) {
                    throw new Error('no padchat manager');
                }
                yield this.padchatManager.WXSetHeadImage(yield file.toBase64());
                return;
            }
            /**
             * 2. get avatar
             */
            const payload = yield this.contactPayload(contactId);
            if (!payload.avatar) {
                throw new Error('no avatar');
            }
            const fileBox = file_box_1.FileBox.fromUrl(payload.avatar, `wechaty-contact-avatar-${payload.name}.jpg`);
            return fileBox;
        });
    }
    contactSelfQrcode() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'contactSelfQrcode()');
            const contactId = this.selfId();
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const contactPayload = yield this.contactPayload(contactId);
            const contactName = contactPayload.alias || contactPayload.name || contactPayload.id;
            return this.getQRCode(this.padchatManager, contactName, contactId);
        });
    }
    getQRCode(manager, contactName, contactId, counter) {
        return __awaiter(this, void 0, void 0, function* () {
            const base64 = yield manager.WXGetUserQRCode(contactId, 3);
            const fileBox = file_box_1.FileBox.fromBase64(base64, `${contactName}.jpg`);
            try {
                // There are some styles of qrcode can not be parsed by the library we are using,
                // So added a retry mechanism here to guarantee the qrcode
                // But still sometimes, the qrcode would be not available
                // So in the error message, let the user to do a retry
                return yield pure_function_helpers_1.fileBoxToQrcode(fileBox);
            }
            catch (e) {
                if (!counter) {
                    counter = 1;
                }
                if (counter > config_1.SELF_QRCODE_MAX_RETRY) {
                    config_1.log.verbose('PuppetPadchat', 'contactQrcode(%s) get qrcode , this should happen very rare', contactId);
                    throw Error('Unable to get qrcode for self, Please try , this issue usually won\'t happen frequently, retry should fix it. If not, please open an issue on https://github.com/lijiarui/wechaty-puppet-padchat');
                }
                return this.getQRCode(manager, contactName, contactId, ++counter);
            }
        });
    }
    contactPayloadDirty(contactId) {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'contactPayloadDirty(%s)', contactId);
            if (this.padchatManager) {
                this.padchatManager.contactRawPayloadDirty(contactId);
            }
            yield _super("contactPayloadDirty").call(this, contactId);
        });
    }
    contactRawPayload(contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PuppetPadchat', 'contactRawPayload(%s)', contactId);
            if (!this.id) {
                throw Error('bot not login!');
            }
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const rawPayload = yield this.padchatManager.contactRawPayload(contactId);
            if (!rawPayload.user_name && contactId === this.id) {
                return pure_function_helpers_1.generateFakeSelfBot(contactId);
            }
            return rawPayload;
        });
    }
    contactRawPayloadParser(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PuppetPadchat', 'contactRawPayloadParser({user_name="%s"})', rawPayload.user_name);
            const payload = pure_function_helpers_1.contactRawPayloadParser(rawPayload);
            if (rawPayload.stranger && pure_function_helpers_1.isStrangerV1(rawPayload.stranger)) {
                payload.friend = true;
            }
            else {
                payload.friend = false;
            }
            // if (!this.padchatManager) {
            //   throw new Error('no padchat manager')
            // }
            // const searchResult = await this.padchatManager.WXSearchContact(rawPayload.user_name)
            // let friend: undefined | boolean = undefined
            // if (searchResult) {
            //   if (searchResult.status === -24 && !searchResult.user_name) {
            //     friend = false
            //   } else if (  isStrangerV1(searchResult.user_name)
            //             || isStrangerV2(searchResult.user_name)
            //   ) {
            //     friend = false
            //   }
            // }
            // return {
            //   ...payload,
            //   friend,
            // }
            return payload;
        });
    }
    /**
     * Overwrite the Puppet.contactPayload()
     */
    contactPayload(contactId) {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const payload = yield _super("contactPayload").call(this, contactId);
                return payload;
            }
            catch (e) {
                config_1.log.silly('PuppetPadchat', 'contactPayload(%s) exception: %s', contactId, e.message);
                config_1.log.silly('PuppetPadchat', 'contactPayload(%s) get failed for %s, try load from room member data source', contactId);
            }
            const rawPayload = yield this.contactRawPayload(contactId);
            /**
             * Issue #1397
             *  https://github.com/Chatie/wechaty/issues/1397#issuecomment-400962638
             *
             * Try to use the contact information from the room
             * when it is not available directly
             */
            if (!rawPayload || Object.keys(rawPayload).length <= 0) {
                config_1.log.silly('PuppetPadchat', 'contactPayload(%s) rawPayload not exist', contactId);
                const roomList = yield this.contactRoomList(contactId);
                config_1.log.silly('PuppetPadchat', 'contactPayload(%s) found %d rooms', contactId, roomList.length);
                if (roomList.length > 0) {
                    const roomId = roomList[0];
                    const roomMemberPayload = yield this.roomMemberPayload(roomId, contactId);
                    if (roomMemberPayload) {
                        const payload = {
                            avatar: roomMemberPayload.avatar,
                            gender: wechaty_puppet_1.ContactGender.Unknown,
                            id: roomMemberPayload.id,
                            name: roomMemberPayload.name,
                            type: wechaty_puppet_1.ContactType.Personal,
                        };
                        this.cacheContactPayload.set(contactId, payload);
                        config_1.log.silly('PuppetPadchat', 'contactPayload(%s) cache SET', contactId);
                        return payload;
                    }
                }
                throw new Error('no raw payload');
            }
            return this.contactRawPayloadParser(rawPayload);
        });
    }
    /**
     *
     * Message
     *
     */
    messageFile(messageId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.warn('PuppetPadchat', 'messageFile(%s)', messageId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const rawPayload = yield this.messageRawPayload(messageId);
            const payload = yield this.messagePayload(messageId);
            const rawText = JSON.stringify(rawPayload);
            const attachmentName = payload.filename || payload.id;
            let result;
            switch (payload.type) {
                case wechaty_puppet_1.MessageType.Audio:
                    return this.getVoiceFileBoxFromRawPayload(rawPayload, attachmentName);
                case wechaty_puppet_1.MessageType.Emoticon:
                    const emojiPayload = yield message_emoji_payload_parser_1.emojiPayloadParser(rawPayload);
                    if (emojiPayload) {
                        return file_box_1.FileBox.fromUrl(emojiPayload.cdnurl, `${attachmentName}.gif`);
                    }
                    else {
                        throw new Error('Can not get emoji file from the message');
                    }
                case wechaty_puppet_1.MessageType.Image:
                    result = yield this.padchatManager.WXGetMsgImage(rawText);
                    return file_box_1.FileBox.fromBase64(result.image, `${attachmentName}.jpg`);
                case wechaty_puppet_1.MessageType.Video:
                    result = yield this.padchatManager.WXGetMsgVideo(rawText);
                    return file_box_1.FileBox.fromBase64(result.video, `${attachmentName}.mp4`);
                case wechaty_puppet_1.MessageType.Attachment:
                default:
                    config_1.log.warn('PuppetPadchat', 'messageFile(%s) unsupport type: %s(%s) because it is not fully implemented yet, PR is welcome.', messageId, padchat_schemas_1.PadchatMessageType[rawPayload.sub_type], rawPayload.sub_type);
                    const base64 = 'Tm90IFN1cHBvcnRlZCBBdHRhY2htZW50IEZpbGUgVHlwZSBpbiBNZXNzYWdlLgpTZWU6IGh0dHBzOi8vZ2l0aHViLmNvbS9DaGF0aWUvd2VjaGF0eS9pc3N1ZXMvMTI0OQo=';
                    const filename = 'wechaty-puppet-padchat-message-attachment-' + messageId + '.txt';
                    const file = file_box_1.FileBox.fromBase64(base64, filename);
                    return file;
            }
        });
    }
    messageUrl(messageId) {
        return __awaiter(this, void 0, void 0, function* () {
            const rawPayload = yield this.messageRawPayload(messageId);
            const payload = yield this.messagePayload(messageId);
            if (payload.type !== wechaty_puppet_1.MessageType.Url) {
                throw new Error('Can not get url from non url payload');
            }
            else {
                const appPayload = yield pure_function_helpers_1.appMessageParser(rawPayload);
                if (appPayload) {
                    return {
                        description: appPayload.des,
                        thumbnailUrl: appPayload.thumburl,
                        title: appPayload.title,
                        url: appPayload.url,
                    };
                }
                else {
                    throw new Error('Can not parse url message payload');
                }
            }
        });
    }
    getVoiceFileBoxFromRawPayload(rawPayload, attachmentName) {
        return __awaiter(this, void 0, void 0, function* () {
            const data = yield this.getVoiceDataFromRawPayload(rawPayload);
            const result = file_box_1.FileBox.fromBase64(data, attachmentName);
            try {
                const match = rawPayload.content.match(/voicelength="(\d+)"/) || [];
                const voiceLength = parseInt(match[1], 10) || 0;
                result.name = `${rawPayload.msg_id}.${voiceLength}.slk`;
            }
            catch (e) {
                config_1.log.error('PuppetPadchat', 'Can not get voice length from content, will have empty voice length');
            }
            return result;
        });
    }
    getVoiceDataFromRawPayload(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            if (!rawPayload.data) {
                const result = yield this.padchatManager.WXGetMsgVoice(JSON.stringify(rawPayload));
                return result.voice;
            }
            else {
                return rawPayload.data;
            }
        });
    }
    messageRawPayload(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const rawPayload = this.cachePadchatMessagePayload.get(id);
            if (!rawPayload) {
                throw new Error('no rawPayload');
            }
            return rawPayload;
        });
    }
    messageRawPayloadParser(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadChat', 'messageRawPayloadParser({msg_id="%s"})', rawPayload.msg_id);
            const payload = yield pure_function_helpers_1.messageRawPayloadParser(rawPayload);
            config_1.log.silly('PuppetPadchat', 'messagePayload(%s)', JSON.stringify(payload));
            return payload;
        });
    }
    messageSendText(receiver, text, mentionIdList) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'messageSend(%s, %s, %s)', JSON.stringify(receiver), text, mentionIdList && mentionIdList.join(','));
            // Send to the Room if there's a roomId
            const id = receiver.roomId || receiver.contactId;
            if (!id) {
                throw Error('no id');
            }
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            yield this.padchatManager.WXSendMsg(id, text, mentionIdList && mentionIdList);
        });
    }
    messageSendFile(receiver, file) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'messageSend("%s", %s)', JSON.stringify(receiver), file);
            // Send to the Room if there's a roomId
            const id = receiver.roomId || receiver.contactId;
            if (!id) {
                throw new Error('no id!');
            }
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const type = file.mimeType || path_1.default.extname(file.name);
            switch (type) {
                case '.slk':
                    try {
                        // TODO: temporary hack solution, replace this when there is metadata in FileBox object
                        const voiceLength = parseInt(file.name.split('.')[1], 10);
                        yield this.padchatManager.WXSendVoice(id, yield file.toBase64(), voiceLength);
                    }
                    catch (e) {
                        throw Error('Can not send voice file, voice length not found from file name, please use voice file generated by wechaty, and don\' modify the file object');
                    }
                    break;
                default:
                    yield this.padchatManager.WXSendImage(id, yield file.toBase64());
                    break;
            }
        });
    }
    messageSendContact(receiver, contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'messageSendContact("%s", %s)', JSON.stringify(receiver), contactId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            // Send to the Room if there's a roomId
            const id = receiver.roomId || receiver.contactId;
            if (!id) {
                throw Error('no id');
            }
            const payload = yield this.contactPayload(contactId);
            const title = payload.name + '名片';
            yield this.padchatManager.WXShareCard(id, contactId, title);
        });
    }
    messageSendUrl(receiver, urlLinkPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'messageSendLink("%s", %s)', JSON.stringify(receiver), JSON.stringify(urlLinkPayload));
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            // Send to the Room if there's a roomId
            const id = receiver.roomId || receiver.contactId;
            if (!id) {
                throw new Error('There is no receiver id when trying to send url link.');
            }
            yield this.padchatManager.WXSendAppMsg(id, app_message_generator_1.generateAppXMLMessage(urlLinkPayload));
        });
    }
    forwardAttachment(receiver, messageId) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const rawPayload = yield this.messageRawPayload(messageId);
            const payload = yield this.messagePayload(messageId);
            // Send to the Room if there's a roomId
            const id = receiver.roomId || receiver.contactId;
            if (!id) {
                throw new Error('There is no receiver id when trying to forward attachment.');
            }
            const appPayload = yield pure_function_helpers_1.appMessageParser(rawPayload);
            if (appPayload === null) {
                throw new Error('Can not forward attachment, failed to parse xml message.');
            }
            const content = app_message_generator_1.generateAttachmentXMLMessageFromRaw(appPayload);
            yield this.padchatManager.WXSendAppMsg(id, content);
        });
    }
    messageForward(receiver, messageId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'messageForward(%s, %s)', JSON.stringify(receiver), messageId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const payload = yield this.messagePayload(messageId);
            if (payload.type === wechaty_puppet_1.MessageType.Text) {
                if (!payload.text) {
                    throw new Error('no text');
                }
                yield this.messageSendText(receiver, payload.text);
            }
            else if (payload.type === wechaty_puppet_1.MessageType.Audio) {
                const rawPayload = yield this.messageRawPayload(messageId);
                const id = receiver.roomId || receiver.contactId;
                if (!id) {
                    throw Error(`Can not find the receiver id for forwarding voice message(${rawPayload.msg_id}), forward voice message failed`);
                }
                let voiceLength;
                try {
                    const match = rawPayload.content.match(/voicelength="(\d+)"/) || [];
                    voiceLength = parseInt(match[1], 10) || 0;
                }
                catch (e) {
                    config_1.log.error(e);
                    throw new Error(`Can not get the length of the voice message(${rawPayload.msg_id}), forward voice message failed`);
                }
                const res = yield this.padchatManager.WXSendVoice(id, yield this.getVoiceDataFromRawPayload(rawPayload), voiceLength);
                config_1.log.error(res);
            }
            else if (payload.type === wechaty_puppet_1.MessageType.Url) {
                yield this.messageSendUrl(receiver, yield this.messageUrl(messageId));
            }
            else if (payload.type === wechaty_puppet_1.MessageType.Attachment) {
                yield this.forwardAttachment(receiver, messageId);
            }
            else {
                yield this.messageSendFile(receiver, yield this.messageFile(messageId));
            }
        });
    }
    /**
     *
     * Room
     *
     */
    roomMemberPayloadDirty(roomId) {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PuppetPadchat', 'roomMemberRawPayloadDirty(%s)', roomId);
            yield _super("roomMemberPayloadDirty").call(this, roomId);
            if (this.padchatManager) {
                yield this.padchatManager.roomMemberRawPayloadDirty(roomId);
            }
        });
    }
    roomMemberRawPayload(roomId, contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PuppetPadchat', 'roomMemberRawPayload(%s)', roomId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const memberDictRawPayload = yield this.padchatManager.roomMemberRawPayload(roomId);
            return memberDictRawPayload[contactId];
        });
    }
    roomMemberRawPayloadParser(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.silly('PuppetPadchat', 'roomMemberRawPayloadParser(%s)', rawPayload);
            const payload = {
                avatar: rawPayload.big_head,
                id: rawPayload.user_name,
                inviterId: rawPayload.invited_by,
                name: rawPayload.nick_name,
                roomAlias: rawPayload.chatroom_nick_name,
            };
            return payload;
        });
    }
    roomPayloadDirty(roomId) {
        const _super = name => super[name];
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomPayloadDirty(%s)', roomId);
            if (this.padchatManager) {
                this.padchatManager.roomRawPayloadDirty(roomId);
            }
            yield _super("roomPayloadDirty").call(this, roomId);
        });
    }
    roomRawPayload(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomRawPayload(%s)', roomId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const rawPayload = yield this.padchatManager.roomRawPayload(roomId);
            if (!rawPayload.user_name)
                rawPayload.user_name = roomId;
            return rawPayload;
        });
    }
    roomRawPayloadParser(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomRawPayloadParser(rawPayload.user_name="%s")', rawPayload.user_name);
            const payload = pure_function_helpers_1.roomRawPayloadParser(rawPayload);
            return payload;
        });
    }
    roomMemberList(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomMemberList(%s)', roomId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const memberIdList = yield this.padchatManager.getRoomMemberIdList(roomId);
            config_1.log.silly('PuppetPadchat', 'roomMemberList()=%d', memberIdList.length);
            if (memberIdList.length <= 0) {
                yield this.roomPayloadDirty(roomId);
            }
            return memberIdList;
        });
    }
    roomValidate(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomValid(%s)', roomId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const exist = yield this.padchatManager.WXGetChatRoomMember(roomId);
            return !!exist;
        });
    }
    roomList() {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomList()');
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const roomIdList = yield this.padchatManager.getRoomIdList();
            config_1.log.silly('PuppetPadchat', 'roomList()=%d', roomIdList.length);
            return roomIdList;
        });
    }
    roomDel(roomId, contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomDel(%s, %s)', roomId, contactId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const memberIdList = yield this.roomMemberList(roomId);
            if (memberIdList.includes(contactId)) {
                yield this.padchatManager.WXDeleteChatRoomMember(roomId, contactId);
            }
            else {
                config_1.log.warn('PuppetPadchat', 'roomDel() room(%s) has no member contact(%s)', roomId, contactId);
            }
            /**
             * Should not dirty payload at here,
             * because later we need to get the leaverId from the event.
             * We will dirty the payload when we process the leave event.
             * Issue #1329 - https://github.com/Chatie/wechaty/issues/1329
             */
            // await this.roomMemberPayloadDirty(roomId)
        });
    }
    roomQrcode(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomQrcode(%s)', roomId);
            const memberIdList = yield this.roomMemberList(roomId);
            if (!memberIdList.includes(this.selfId())) {
                throw new Error('userSelf not in this room: ' + roomId);
            }
            const base64 = yield this.padchatManager.WXGetUserQRCode(roomId, 0);
            const roomPayload = yield this.roomPayload(roomId);
            const roomName = roomPayload.topic || roomPayload.id;
            const fileBox = file_box_1.FileBox.fromBase64(base64, `${roomName}-qrcode.jpg`);
            const qrcode = yield pure_function_helpers_1.fileBoxToQrcode(fileBox);
            return qrcode;
        });
    }
    roomAvatar(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomAvatar(%s)', roomId);
            const payload = yield this.roomPayload(roomId);
            if (payload.avatar) {
                return file_box_1.FileBox.fromUrl(payload.avatar);
            }
            config_1.log.warn('PuppetPadchat', 'roomAvatar() avatar not found, use the chatie default.');
            return config_1.qrCodeForChatie();
        });
    }
    roomAdd(roomId, contactId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomAdd(%s, %s)', roomId, contactId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            // XXX: did there need to calc the total number of the members in this room?
            // if n <= 40 then add() else invite() ?
            try {
                config_1.log.verbose('PuppetPadchat', 'roomAdd(%s, %s) try to Add', roomId, contactId);
                yield this.padchatManager.WXAddChatRoomMember(roomId, contactId);
            }
            catch (e) {
                // FIXME
                console.error(e);
                config_1.log.warn('PuppetPadchat', 'roomAdd(%s, %s) Add exception: %s', e);
                config_1.log.verbose('PuppetPadchat', 'roomAdd(%s, %s) try to Invite', roomId, contactId);
                yield this.padchatManager.WXInviteChatRoomMember(roomId, contactId);
            }
            // Will reload room information when receive room-join event
            // instead of here, since the room information might not be updated yet
            // await new Promise(r => setTimeout(r, 1000))
            // await this.roomMemberPayloadDirty(roomId)
            // await this.roomMemberPayload(roomId, contactId)
        });
    }
    roomTopic(roomId, topic) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomTopic(%s, %s)', roomId, topic);
            if (typeof topic === 'undefined') {
                const payload = yield this.roomPayload(roomId);
                return payload.topic;
            }
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            yield this.padchatManager.WXSetChatroomName(roomId, topic);
            /**
             * Give server some time to refresh the API payload
             * when we have to make sure the data is the latest.
             */
            yield new Promise(r => setTimeout(r, 1000));
            yield this.roomPayloadDirty(roomId);
            yield this.roomPayload(roomId);
            return;
        });
    }
    roomCreate(contactIdList, topic) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomCreate(%s, %s)', contactIdList, topic);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            const roomId = yield this.padchatManager.WXCreateChatRoom(contactIdList);
            // Load new created room payload
            yield this.roomPayload(roomId);
            return roomId;
        });
    }
    roomQuit(roomId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomQuit(%s)', roomId);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            yield this.padchatManager.WXQuitChatRoom(roomId);
            // Clean Cache
            yield this.roomMemberPayloadDirty(roomId);
            yield this.roomPayloadDirty(roomId);
        });
    }
    roomAnnounce(roomId, text) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'roomAnnounce(%s, %s)', roomId, text ? text : '');
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            if (text) {
                yield this.padchatManager.WXSetChatroomAnnouncement(roomId, text);
            }
            else {
                config_1.log.warn('Getting room announcement is not supported by wechaty-puppet-padchat.');
                return '';
            }
        });
    }
    roomInvitationRawPayload(roomInvitationId) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            return this.padchatManager.roomInvitationRawPayload(roomInvitationId);
        });
    }
    roomInvitationRawPayloadParser(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            return {
                id: rawPayload.id,
                inviterId: rawPayload.fromUser,
                roomMemberCount: 0,
                roomMemberIdList: [],
                roomTopic: rawPayload.roomName,
                timestamp: rawPayload.timestamp
            };
        });
    }
    roomInvitationAccept(roomInvitationId) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.padchatManager) {
                throw new Error('no padcaht manager');
            }
            let res = '';
            try {
                const payload = yield this.padchatManager.roomInvitationRawPayload(roomInvitationId);
                const shareUrl = payload.url;
                const response = yield this.padchatManager.WXGetRequestToken(this.selfId(), shareUrl);
                res = yield require('request-promise')({
                    method: 'POST',
                    simple: false,
                    uri: response.full_url,
                });
            }
            catch (e) {
                throw new Error('UNKNOWN: Unexpected error happened when trying to accept invitation\n' + e);
            }
            if (res.indexOf('你无法查看被转发过的邀请') !== -1 || res.indexOf('Unable to view forwarded invitations') === -1) {
                throw new Error('FORWARDED: Accept invitation failed, this is a forwarded invitation, can not be accepted');
            }
            else if (res.indexOf('你未开通微信支付') !== -1 || res.indexOf('You haven\'t enabled WeChat Pay') === -1
                || res.indexOf('你需要实名验证后才能接受邀请') !== -1) {
                throw new Error('WXPAY: The user need to enable wechaty pay(微信支付) to join the room, this is requested by Wechat.');
            }
            else if (res.indexOf('该邀请已过期') !== -1 || res.indexOf('Invitation expired') === -1) {
                throw new Error('EXPIRED: The invitation is expired, please request the user to send again');
            }
            else if (res.indexOf('群聊邀请操作太频繁请稍后再试') !== -1 || res.indexOf('操作太频繁，请稍后再试') !== -1) {
                throw new Error('FREQUENT: Room invitation operation too frequent.');
            }
            else if (res.indexOf('已达群聊人数上限') !== -1) {
                throw new Error('LIMIT: The room member count has reached the limit.');
            }
            else if (res.indexOf('该群因违规已被限制使用，无法添加群成员') !== -1) {
                throw new Error('INVALID: This room has been mal used, can not add new members.');
            }
        });
    }
    /**
     *
     * Friendship
     *
     */
    friendshipAdd(contactId, hello) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'friendshipAdd(%s, %s)', contactId, hello);
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            let rawSearchPayload;
            try {
                rawSearchPayload = yield this.padchatManager.WXSearchContact(contactId);
            }
            catch (e) {
                throw Error(`Can not add user ${contactId}, this contactId is not searchable. Please refer to issue: https://github.com/lijiarui/wechaty-puppet-padchat/issues/166`);
            }
            /**
             * If the contact is not stranger, than ussing WXSearchContact can get user_name
             */
            if (rawSearchPayload.user_name !== '' && !pure_function_helpers_1.isStrangerV1(rawSearchPayload.user_name) && !pure_function_helpers_1.isStrangerV2(rawSearchPayload.user_name)) {
                config_1.log.warn('PuppetPadchat', 'friendshipAdd %s has been friend with bot, no need to send friend request!', contactId);
                return;
            }
            let strangerV1;
            let strangerV2;
            if (pure_function_helpers_1.isStrangerV1(rawSearchPayload.stranger)) {
                strangerV1 = rawSearchPayload.stranger;
                strangerV2 = rawSearchPayload.user_name;
            }
            else if (pure_function_helpers_1.isStrangerV2(rawSearchPayload.stranger)) {
                strangerV2 = rawSearchPayload.stranger;
                strangerV1 = rawSearchPayload.user_name;
            }
            else {
                throw new Error('stranger neither v1 nor v2!');
            }
            // Issue #1252 : what's wrong here?, Trying to fix now...
            const isPhoneNumber = contactId.match(/^(?=\d{11}$)^1(?:3\d|4[57]|5[^4\D]|66|7[^249\D]|8\d|9[89])\d{8}$/);
            yield this.padchatManager.WXAddUser(strangerV1 || '', strangerV2 || '', isPhoneNumber ? padchat_rpc_type_1.WXSearchContactTypeStatus.MOBILE : padchat_rpc_type_1.WXSearchContactTypeStatus.WXID, // default
            hello);
        });
    }
    friendshipAccept(friendshipId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'friendshipAccept(%s)', friendshipId);
            const payload = yield this.friendshipPayload(friendshipId);
            if (!payload.ticket) {
                throw new Error('no ticket');
            }
            if (!payload.stranger) {
                throw new Error('no stranger');
            }
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            yield this.padchatManager.WXAcceptUser(payload.stranger, payload.ticket);
        });
    }
    friendshipRawPayloadParser(rawPayload) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'friendshipRawPayloadParser({id=%s})', rawPayload.msg_id);
            const payload = yield pure_function_helpers_1.friendshipRawPayloadParser(rawPayload);
            return payload;
        });
    }
    friendshipRawPayload(friendshipId) {
        return __awaiter(this, void 0, void 0, function* () {
            config_1.log.verbose('PuppetPadchat', 'friendshipRawPayload(%s)', friendshipId);
            /**
             * Friendship shares Cache with the Message RawPayload
             */
            const rawPayload = this.cachePadchatMessagePayload.get(friendshipId);
            if (!rawPayload) {
                throw new Error('no rawPayload for id ' + friendshipId);
            }
            return rawPayload;
        });
    }
    unref() {
        config_1.log.verbose('PuppetPadchat', 'unref ()');
        super.unref();
        if (this.padchatManager) {
            // TODO: this.padchatManager.unref()
        }
    }
    contactSelfName(newName) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            yield this.padchatManager.updateSelfName(newName);
            yield this.contactPayloadDirty(this.selfId());
        });
    }
    contactSelfSignature(signature) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.padchatManager) {
                throw new Error('no padchat manager');
            }
            yield this.padchatManager.updateSelfSignature(signature);
            yield this.contactPayloadDirty(this.selfId());
        });
    }
}
PuppetPadchat.VERSION = config_1.VERSION;
exports.PuppetPadchat = PuppetPadchat;
exports.default = PuppetPadchat;
//# sourceMappingURL=puppet-padchat.js.map